@@ -26,6 +26,8 @@ import sys
 import tempfile
 
 from cx_Freeze import Executable, setup  # pylint: disable=import-error
+from cx_Freeze.hooks import gi  # pylint: disable=import-error
+del gi.load_gi
 
 # pylint: disable=duplicate-code
 
@@ -45,7 +47,7 @@ elif sys.platform == "darwin":
     SYS_BASE_PATH = "/opt/homebrew" if platform.machine() == "arm64" else "/usr/local"
     LIB_PATH = os.path.join(SYS_BASE_PATH, "lib")
     LIB_EXTENSION = (".dylib", ".so")
-    UNAVAILABLE_MODULES = ["msilib", "msvcrt", "nt", "nturl2path", "ossaudiodev", "spwd", "winreg", "winsound"]
+    UNAVAILABLE_MODULES = ["msvcrt", "nt", "nturl2path", "ossaudiodev", "spwd", "winreg", "winsound"]
     ICON_NAME = "icon.icns"
 
 else:
@@ -61,12 +63,13 @@ import pynicotine  # noqa: E402  # pylint: disable=import-error,wrong-import-pos
 
 SCRIPT_NAME = "nicotine"
 MODULE_NAME = "pynicotine"
-GTK_VERSION = os.environ.get("NICOTINE_GTK_VERSION", "4")
-USE_LIBADWAITA = GTK_VERSION == "4" and os.environ.get("NICOTINE_LIBADWAITA") == "1"
+MANIFEST_NAME = os.path.join(CURRENT_PATH, f"{SCRIPT_NAME}.manifest") if sys.platform == "win32" else None
 
 # Include (almost) all standard library modules for plugins
 EXCLUDED_MODULES = UNAVAILABLE_MODULES + [
-    "ensurepip", "idlelib", "pip", "tkinter", "turtle", "turtledemo", "venv", "zoneinfo"
+    f"{MODULE_NAME}.plugins.examplars", f"{MODULE_NAME}.tests",
+    "ctypes.test", "distutils", "ensurepip", "idlelib", "lib2to3", "msilib", "pip", "pydoc", "pydoc_data",
+    "pygtkcompat", "tkinter", "turtle", "turtledemo", "unittest.test", "venv", "zoneinfo"
 ]
 INCLUDED_MODULES = [MODULE_NAME, "gi"] + list(
     # pylint: disable=no-member
@@ -150,35 +153,26 @@ def _add_typelibs_callback(full_path, short_path, _callback_data=None):
 def add_typelibs():
 
     required_typelibs = [
-        f"Gtk-{GTK_VERSION}",
+        "Adw-",
+        "Gtk-4",
         "Gio-",
-        f"Gdk-{GTK_VERSION}",
-        f"GdkWin32-{GTK_VERSION}",
+        "Gdk-4",
         "GLib-",
+        "Graphene-",
+        "Gsk-",
         "HarfBuzz-",
         "Pango-",
+        "PangoCairo-",
         "GObject-",
         "GdkPixbuf-",
         "cairo-",
         "GModule-",
-        "freetype2-",
-        "win32-"
+        "freetype2-"
     ]
 
-    if GTK_VERSION == "4":
-        required_typelibs += [
-            "Graphene-",
-            "Gsk-",
-            "PangoCairo-"
-        ]
-    else:
-        required_typelibs += [
-            "Atk-",
-            "Gspell-"
-        ]
-
-    if USE_LIBADWAITA:
-        required_typelibs.append("Adw-")
+    if sys.platform == "win32":
+        required_typelibs.append("GdkWin32-4")
+        required_typelibs.append("win32-")
 
     required_typelibs = tuple(required_typelibs)
     folder_path = os.path.join(SYS_BASE_PATH, "lib/girepository-1.0")
@@ -206,14 +200,12 @@ def add_gtk():
     # This also includes all dlls required by GTK
     add_files(
         folder_path=LIB_PATH, output_path="lib",
-        starts_with=f"libgtk-{GTK_VERSION}", ends_with=LIB_EXTENSION
+        starts_with="libgtk-4", ends_with=LIB_EXTENSION
+    )
+    add_files(
+        folder_path=LIB_PATH, output_path="lib",
+        starts_with="libadwaita-", ends_with=LIB_EXTENSION
     )
-
-    if USE_LIBADWAITA:
-        add_files(
-            folder_path=LIB_PATH, output_path="lib",
-            starts_with="libadwaita-", ends_with=LIB_EXTENSION
-        )
 
     # Schemas
     add_file(
@@ -245,19 +237,6 @@ def add_icon_packs():
     )
 
 
-def add_themes():
-
-    # "Mac" is required for macOS-specific keybindings in GTK
-    required_themes = (
-        "Default",
-        "Mac"
-    )
-    add_files(
-        folder_path=os.path.join(SYS_BASE_PATH, "share/themes"), output_path="share/themes",
-        starts_with=required_themes, ends_with=".css", recursive=True
-    )
-
-
 def add_ssl_certs():
     ssl_paths = ssl.get_default_verify_paths()
     add_file(file_path=ssl_paths.openssl_cafile, output_path="lib/cert.pem")
@@ -270,14 +249,13 @@ def add_translations():
 
     add_files(
         folder_path=os.path.join(SYS_BASE_PATH, "share/locale"), output_path="share/locale",
-        starts_with=tuple(i[0] for i in pynicotine.i18n.LANGUAGES), ends_with=f"gtk{GTK_VERSION}0.mo", recursive=True
+        starts_with=tuple(i[0] for i in pynicotine.i18n.LANGUAGES), ends_with="gtk40.mo", recursive=True
     )
 
 
 # GTK
 add_gtk()
 add_icon_packs()
-add_themes()
 
 # SSL
 add_ssl_certs()
@@ -301,7 +279,8 @@ setup(
             "excludes": EXCLUDED_MODULES,
             "include_files": include_files,
             "zip_include_packages": ["*"],
-            "zip_exclude_packages": [MODULE_NAME]
+            "zip_exclude_packages": [MODULE_NAME],
+            "optimize": 2
         },
         "bdist_msi": {
             "all_users": True,
@@ -340,6 +319,7 @@ setup(
             base=GUI_BASE,
             target_name=pynicotine.__application_name__,
             icon=os.path.join(CURRENT_PATH, ICON_NAME),
+            manifest=MANIFEST_NAME,
             copyright=pynicotine.__copyright__,
             shortcut_name=pynicotine.__application_name__,
             shortcut_dir="ProgramMenuFolder"
