@@ -1,4 +1,4 @@
-# COPYRIGHT (C) 2020-2023 Nicotine+ Contributors
+# COPYRIGHT (C) 2020-2024 Nicotine+ Contributors
 # COPYRIGHT (C) 2016-2017 Michael Labouebe <gfarmerfr@free.fr>
 # COPYRIGHT (C) 2016-2018 Mutnick <mutnick@techie.com>
 # COPYRIGHT (C) 2008-2011 quinox <quinox@users.sf.net>
@@ -49,6 +49,9 @@ class Config:
     parameters.
     """
 
+    __slots__ = ("config_file_path", "data_folder_path", "config_loaded", "sections",
+                 "defaults", "removed_options", "_parser")
+
     def __init__(self):
 
         config_folder_path, data_folder_path = self.get_user_folders()
@@ -81,7 +84,7 @@ class Config:
         home = os.path.expanduser("~")
         legacy_folder_path = os.path.join(home, ".nicotine")
 
-        if os.path.isdir(legacy_folder_path.encode("utf-8")):
+        if os.path.isdir(encode_path(legacy_folder_path)):
             return legacy_folder_path, legacy_folder_path
 
         def xdg_path(xdg, default):
@@ -149,8 +152,10 @@ class Config:
 
         from configparser import ConfigParser
 
+        data_home_env = "${NICOTINE_DATA_HOME}"
+        log_folder_path = os.path.join(data_home_env, "logs")
+
         self._parser = ConfigParser(strict=False, interpolation=None)
-        log_folder_path = os.path.join("${NICOTINE_DATA_HOME}", "logs")
         self.defaults = {
             "server": {
                 "server": ("server.slsknet.org", 2242),
@@ -169,16 +174,15 @@ class Config:
                 "ignorelist": [],
                 "ipignorelist": {},
                 "ipblocklist": {},
-                "autojoin": ["nicotine"],
+                "autojoin": [],
                 "autoaway": 15,
                 "away": False,
-                "private_chatrooms": False,
-                "command_aliases": {}
+                "private_chatrooms": False
             },
             "transfers": {
-                "incompletedir": os.path.join("${NICOTINE_DATA_HOME}", "incomplete"),
-                "downloaddir": os.path.join("${NICOTINE_DATA_HOME}", "downloads"),
-                "uploaddir": os.path.join("${NICOTINE_DATA_HOME}", "received"),
+                "incompletedir": os.path.join(data_home_env, "incomplete"),
+                "downloaddir": os.path.join(data_home_env, "downloads"),
+                "uploaddir": os.path.join(data_home_env, "received"),
                 "usernamesubfolders": False,
                 "shared": [],
                 "buddyshared": [],
@@ -218,10 +222,11 @@ class Config:
                 "enablefilters": False,
                 "downloadregexp": "",
                 "downloadfilters": [
+                    ["*.DS_Store", 1],
+                    ["*.exe", 1],
+                    ["*.msi", 1],
                     ["desktop.ini", 1],
-                    ["*.url", 1],
-                    ["thumbs.db", 1],
-                    ["albumart(_{........-....-....-....-............}_)?(_?(large|small))?\\.jpg", 0]
+                    ["Thumbs.db", 1]
                 ],
                 "download_doubleclick": 2,
                 "upload_doubleclick": 2,
@@ -271,7 +276,6 @@ class Config:
                 "debug_file_output": False,
                 "roomlogsdir": os.path.join(log_folder_path, "rooms"),
                 "privatelogsdir": os.path.join(log_folder_path, "private"),
-                "readroomlogs": True,
                 "readroomlines": 200,
                 "readprivatelines": 200,
                 "private_chats": [],
@@ -292,7 +296,7 @@ class Config:
             "searches": {
                 "expand_searches": True,
                 "group_searches": "folder_grouping",
-                "maxresults": 150,
+                "maxresults": 300,
                 "enable_history": True,
                 "history": [],
                 "enablefilters": False,
@@ -306,9 +310,9 @@ class Config:
                 "filtertype": [],
                 "filterlength": [],
                 "search_results": True,
-                "max_displayed_results": 1500,
+                "max_displayed_results": 2500,
                 "min_search_chars": 3,
-                "private_search_results": True
+                "private_search_results": False
             },
             "ui": {
                 "language": "",
@@ -328,7 +332,6 @@ class Config:
                 "usernamestyle": "bold",
                 "textbg": "",
                 "search": "",
-                "searchq": "GREY",
                 "inputcolor": "",
                 "spellcheck": True,
                 "exitdialog": 1,
@@ -388,9 +391,6 @@ class Config:
                 "reverse_file_paths": True,
                 "file_size_unit": ""
             },
-            "private_rooms": {
-                "rooms": {}
-            },
             "urls": {
                 "protocols": {}
             },
@@ -456,7 +456,8 @@ class Config:
                 "fallbackencodings",
                 "roomencoding",
                 "userencoding",
-                "firewalled"
+                "firewalled",
+                "command_aliases"
             ),
             "ui": (
                 "enabletrans",
@@ -488,7 +489,8 @@ class Config:
                 "decimalsep",
                 "urgencyhint",
                 "tab_status_icons",
-                "file_path_tooltips"
+                "file_path_tooltips",
+                "searchq"
             ),
             "columns": (
                 "downloads",
@@ -526,10 +528,12 @@ class Config:
             ),
             "private_rooms": (
                 "enabled",
+                "rooms"
             ),
             "logging": (
                 "logsdir",
-                "timestamps"
+                "timestamps",
+                "readroomlogs"
             ),
             "ticker": (
                 "default",
@@ -574,10 +578,14 @@ class Config:
         from pynicotine.logfacility import log
         log.init_log_levels()
         log.update_folder_paths()
-        log.add_debug("Using configuration: %(file)s", {"file": self.config_file_path})
+        log.add_debug("Using configuration: %s", self.config_file_path)
 
         events.connect("quit", self._quit)
 
+    def need_config(self):
+        # Check if we have specified a username or password
+        return not self.sections["server"]["login"] or not self.sections["server"]["passw"]
+
     def _parse_config(self, file_path):
         """Parses the config file."""
 
@@ -585,13 +593,69 @@ class Config:
             file_handle.seek(0)
             self._parser.read_file(file_handle)
 
-    def need_config(self):
+    def _migrate_config(self):
 
-        # Check if we have specified a username or password
-        if not self.sections["server"]["login"] or not self.sections["server"]["passw"]:
-            return True
+        # Map legacy folder/user grouping modes (3.1.0)
+        for section, option in (
+            ("searches", "group_searches"),
+            ("transfers", "groupdownloads"),
+            ("transfers", "groupuploads")
+        ):
+            mode = self.sections[section].get(option, "folder_grouping")
 
-        return False
+            if mode == "0":
+                mode = "ungrouped"
+
+            elif mode == "1":
+                mode = "folder_grouping"
+
+            elif mode == "2":
+                mode = "user_grouping"
+
+            self.sections[section][option] = mode
+
+        # Convert special download folder share to regular share
+        if self.sections["transfers"].get("sharedownloaddir", False):
+            shares = self.sections["transfers"]["shared"]
+            virtual_name = "Downloaded"
+            shared_folder = (virtual_name, self.sections["transfers"]["downloaddir"])
+
+            if shared_folder not in shares and virtual_name not in (x[0] for x in shares):
+                shares.append(shared_folder)
+
+        # Migrate download/upload speed limit preference (3.3.0)
+        if "uselimit" in self.sections["transfers"] or "usealtlimits" in self.sections["transfers"]:
+            for option, use_primary_speed_limit in (
+                ("use_download_speed_limit", self.sections["transfers"].get("downloadlimit", 0) > 0),
+                ("use_upload_speed_limit", self.sections["transfers"].get("uselimit", False))
+            ):
+                if self.sections["transfers"].get("usealtlimits", False):
+                    use_speed_limit = "alternative"
+
+                elif use_primary_speed_limit:
+                    use_speed_limit = "primary"
+
+                else:
+                    use_speed_limit = "unlimited"
+
+                self.sections["transfers"][option] = use_speed_limit
+
+        # Migrate old trusted buddy shares to new format (3.3.0)
+        if self.sections["transfers"].get("buddysharestrustedonly", False):
+            buddy_shares = self.sections["transfers"]["buddyshared"]
+
+            self.sections["transfers"]["trustedshared"] = buddy_shares[:]
+            buddy_shares.clear()
+
+        # Migrate old media player command to new format (3.3.0)
+        old_default_player = self.sections["players"].get("default", None)
+
+        if old_default_player:
+            self.sections["urls"]["protocols"]["audio"] = old_default_player
+
+        # Enable previously disabled header bar on macOS (3.3.0)
+        if sys.platform == "darwin" and old_default_player is not None:
+            self.sections["ui"]["header_bar"] = True
 
     def _set_config(self):
         """Set config values parsed from file earlier."""
@@ -609,8 +673,7 @@ class Config:
                 # Check if config option exists in defaults
                 elif (j not in self.defaults.get(i, {}) and j not in self.removed_options.get(i, {})
                         and i != "plugins" and j != "filter"):
-                    log.add_debug("Unknown config option '%(option)s' in section '%(section)s'",
-                                  {"option": j, "section": i})
+                    log.add_debug("Unknown config option '%s' in section '%s'", (j, i))
 
                 else:
                     # Attempt to get the default value for a config option. If there's no default
@@ -668,65 +731,11 @@ class Config:
                 self.sections[section] = {}
 
             for option, value in options.items():
-                if option in self.sections[section]:
-                    continue
+                if option not in self.sections[section]:
+                    self.sections[section][option] = value
 
-                # Migrate download speed limit preference (3.3.0)
-                if option == "use_download_speed_limit" and section == "transfers":
-                    if self.sections[section].get("usealtlimits", False):
-                        use_speed_limit = "alternative"
-
-                    elif self.sections[section].get("downloadlimit", 0) > 0:
-                        use_speed_limit = "primary"
-
-                    else:
-                        use_speed_limit = "unlimited"
-
-                    self.sections[section][option] = use_speed_limit
-                    continue
-
-                # Migrate upload speed limit preference (3.3.0)
-                if option == "use_upload_speed_limit" and section == "transfers":
-                    if self.sections[section].get("usealtlimits", False):
-                        use_speed_limit = "alternative"
-
-                    elif self.sections[section].get("uselimit", False):
-                        use_speed_limit = "primary"
-
-                    else:
-                        use_speed_limit = "unlimited"
-
-                    self.sections[section][option] = use_speed_limit
-                    continue
-
-                # Set default value
-                self.sections[section][option] = value
-
-        # Convert special download folder share to regular share
-        if self.sections["transfers"].get("sharedownloaddir", False):
-            shares = self.sections["transfers"]["shared"]
-            virtual_name = "Downloaded"
-            shared_folder = (virtual_name, self.sections["transfers"]["downloaddir"])
-
-            if shared_folder not in shares and virtual_name not in (x[0] for x in shares):
-                shares.append(shared_folder)
-
-        # Migrate old trusted buddy shares to new format (3.3.0)
-        if self.sections["transfers"].get("buddysharestrustedonly", False):
-            buddy_shares = self.sections["transfers"]["buddyshared"]
-
-            self.sections["transfers"]["trustedshared"] = buddy_shares[:]
-            buddy_shares.clear()
-
-        # Migrate old media player command to new format (3.3.0)
-        old_default_player = self.sections["players"].get("default", None)
-
-        if old_default_player:
-            self.sections["urls"]["protocols"]["audio"] = old_default_player
-
-        # Enable previously disabled header bar on macOS (3.3.0)
-        if sys.platform == "darwin" and old_default_player is not None:
-            self.sections["ui"]["header_bar"] = True
+        # Migrate old config values
+        self._migrate_config()
 
         # Check if server value is valid
         server_addr = self.sections["server"]["server"]
@@ -769,13 +778,16 @@ class Config:
             for option in options:
                 self._parser.remove_option(section, option)
 
+            if not self._parser.options(section):
+                self._parser.remove_section(section)
+
         if not self.create_config_folder():
             return
 
         from pynicotine.logfacility import log
 
         write_file_and_backup(self.config_file_path, self._write_config_callback, protect=True)
-        log.add_debug("Saved configuration: %(file)s", {"file": self.config_file_path})
+        log.add_debug("Saved configuration: %s", self.config_file_path)
 
     def write_config_backup(self, file_path):
 
@@ -808,10 +820,6 @@ class Config:
         if self._parser is not None:
             self._parser.clear()
 
-        self.sections.clear()
-        self.defaults.clear()
-        self.removed_options.clear()
-
         self.config_loaded = False
 
 
