@@ -28,7 +28,6 @@ from gi.repository import GLib
 from gi.repository import Gtk
 
 import pynicotine
-from pynicotine import slskmessages
 from pynicotine.config import config
 from pynicotine.core import core
 from pynicotine.events import events
@@ -94,6 +93,7 @@ class Application:
         for event_name, callback in (
             ("confirm-quit", self.on_confirm_quit),
             ("invalid-password", self.on_invalid_password),
+            ("invalid-username", self.on_invalid_password),
             ("quit", self._instance.quit),
             ("server-login", self._update_user_status),
             ("server-disconnect", self._update_user_status),
@@ -133,13 +133,13 @@ class Application:
             ("connect", self.on_connect, None, True),
             ("disconnect", self.on_disconnect, None, False),
             ("soulseek-privileges", self.on_soulseek_privileges, None, False),
-            ("away", self.on_away, None, False),
+            ("away", self.on_away, None, True),
             ("away-accel", self.on_away_accelerator, None, False),
             ("message-downloading-users", self.on_message_downloading_users, None, False),
             ("message-buddies", self.on_message_buddies, None, False),
             ("wishlist", self.on_wishlist, None, True),
             ("confirm-quit", self.on_confirm_quit_request, None, True),
-            ("confirm-quit-uploads", self.on_confirm_quit_uploads_request, None, True),
+            ("force-quit", self.on_force_quit_request, None, True),
             ("quit", self.on_quit_request, None, True),
 
             # Shares
@@ -222,10 +222,11 @@ class Application:
             # Global accelerators
             ("app.connect", ["<Shift><Primary>c"]),
             ("app.disconnect", ["<Shift><Primary>d"]),
-            ("app.away-accel", ["<Primary>h"]),
+            ("app.away-accel", ["<Shift><Primary>a"]),
             ("app.wishlist", ["<Shift><Primary>w"]),
             ("app.confirm-quit", ["<Primary>q"]),
-            ("app.quit", ["<Primary><Alt>q"]),
+            ("app.force-quit", ["<Primary><Alt>q"]),
+            ("app.quit", ["<Primary>q"]),  # Only used to show accelerator in menus
             ("app.rescan-shares", ["<Shift><Primary>r"]),
             ("app.keyboard-shortcuts", ["<Primary>question", "F1"]),
             ("app.preferences", ["<Primary>comma", "<Primary>p"]),
@@ -270,53 +271,6 @@ class Application:
         # Disable Alt+1-9 accelerators for numpad keys to avoid conflict with Alt codes
         self._set_accels_for_action("app.disabled", numpad_accels)
 
-        if GTK_API_VERSION == 3 or sys.platform != "darwin":
-            return
-
-        # Built-in GTK shortcuts use Ctrl key on macOS, add shortcuts that use Command key
-        for widget in (Gtk.Text, Gtk.TextView):
-            for action_name, accelerator in (
-                ("clipboard.cut", "<Meta>x"),
-                ("clipboard.copy", "<Meta>c"),
-                ("clipboard.paste", "<Meta>v"),
-                ("selection.select-all", "<Meta>a"),
-                ("misc.insert-emoji", "<Meta>period"),
-                ("text.undo", "<Meta>z"),
-                ("text.redo", "<Shift><Meta>u")
-            ):
-                widget.add_shortcut(
-                    Gtk.Shortcut(
-                        trigger=Gtk.ShortcutTrigger.parse_string(accelerator),
-                        action=Gtk.NamedAction(action_name=action_name),
-                    )
-                )
-
-            for accelerator, step, count, extend in (
-                ("<Meta>Up|<Meta>KP_Up", Gtk.MovementStep.BUFFER_ENDS, -1, False),
-                ("<Shift><Meta>Up|<Shift><Meta>KP_Up", Gtk.MovementStep.BUFFER_ENDS, -1, True),
-                ("<Meta>Down|<Meta>KP_Down", Gtk.MovementStep.BUFFER_ENDS, 1, False),
-                ("<Shift><Meta>Down|<Shift><Meta>KP_Down", Gtk.MovementStep.BUFFER_ENDS, 1, True),
-                ("<Meta>Left|<Meta>KP_Left", Gtk.MovementStep.DISPLAY_LINE_ENDS, -1, False),
-                ("<Shift><Meta>Left|<Shift><Meta>KP_Left", Gtk.MovementStep.DISPLAY_LINE_ENDS, -1, True),
-                ("<Meta>Right|<Meta>KP_Right", Gtk.MovementStep.DISPLAY_LINE_ENDS, 1, False),
-                ("<Shift><Meta>Right|<Shift><Meta>KP_Right", Gtk.MovementStep.DISPLAY_LINE_ENDS, 1, True),
-                ("<Alt>Left|<Alt>KP_Left", Gtk.MovementStep.WORDS, -1, False),
-                ("<Shift><Alt>Left|<Shift><Alt>KP_Left", Gtk.MovementStep.WORDS, -1, True),
-                ("<Alt>Right|<Alt>KP_Right", Gtk.MovementStep.WORDS, 1, False),
-                ("<Shift><Alt>Right|<Shift><Alt>KP_Right", Gtk.MovementStep.WORDS, 1, True)
-            ):
-                widget.add_shortcut(
-                    Gtk.Shortcut(
-                        trigger=Gtk.ShortcutTrigger.parse_string(accelerator),
-                        action=Gtk.SignalAction(signal_name="move-cursor"),
-                        arguments=GLib.Variant.new_tuple(
-                            GLib.Variant.new_int32(step),
-                            GLib.Variant.new_int32(count),
-                            GLib.Variant.new_boolean(extend)
-                        )
-                    )
-                )
-
     def _update_user_status(self, *_args):
 
         status = core.users.login_status
@@ -324,11 +278,11 @@ class Application:
 
         self.lookup_action("connect").set_enabled(not is_online)
 
-        for action_name in ("disconnect", "soulseek-privileges", "away-accel", "away",
+        for action_name in ("disconnect", "soulseek-privileges", "away-accel",
                             "message-downloading-users", "message-buddies"):
             self.lookup_action(action_name).set_enabled(is_online)
 
-        self.tray_icon.update_user_status()
+        self.tray_icon.update()
 
     # Primary Menus #
 
@@ -344,13 +298,13 @@ class Application:
 
     @staticmethod
     def _add_preferences_item(menu):
-        menu.add_items(("#" + _("_Preferences"), "app.preferences"))
+        menu.add_items(("^" + _("_Preferences"), "app.preferences"))
 
     def _add_quit_item(self, menu):
 
         menu.add_items(
             ("", None),
-            ("#" + _("_Quit"), "app.confirm-quit-uploads")
+            ("^" + _("_Quit"), "app.quit")
         )
 
     def _create_file_menu(self):
@@ -408,7 +362,7 @@ class Application:
             ("#" + _("Report a _Bug"), "app.report-bug"),
             ("#" + _("Improve T_ranslations"), "app.improve-translations"),
             ("", None),
-            ("#" + _("_About Nicotine+"), "app.about")
+            ("^" + _("_About Nicotine+"), "app.about")
         )
 
         return menu
@@ -458,7 +412,10 @@ class Application:
 
         try:
             if sys.platform == "win32":
-                self.tray_icon.show_notification(title=title, message=message)
+                self.tray_icon.show_notification(
+                    title=title, message=message, action=action, action_target=action_target,
+                    high_priority=high_priority
+                )
                 return
 
             priority = Gio.NotificationPriority.HIGH if high_priority else Gio.NotificationPriority.NORMAL
@@ -475,7 +432,7 @@ class Application:
 
             # Unity doesn't support default click actions, and replaces the notification with a dialog.
             # Disable actions to prevent this from happening.
-            if action and os.environ.get("XDG_SESSION_DESKTOP") != "unity":
+            if action and os.environ.get("XDG_CURRENT_DESKTOP", "").lower() != "unity":
                 if action_target:
                     notification.set_default_action_and_target(action, GLib.Variant("s", action_target))
                 else:
@@ -487,24 +444,39 @@ class Application:
                 Gdk.Display.get_default().beep()
 
         except Exception as error:
-            log.add(_("Unable to show notification: %s"), str(error))
+            log.add(_("Unable to show notification: %s"), error)
 
     def _show_chatroom_notification(self, room, message, title=None, high_priority=False):
+
         self._show_notification(
             message, title, action="app.chatroom-notification-activated", action_target=room,
-            high_priority=high_priority)
+            high_priority=high_priority
+        )
+
+        if high_priority:
+            self.window.set_urgency_hint(True)
 
     def _show_download_notification(self, message, title=None, high_priority=False):
+
         self._show_notification(
-            message, title, action="app.download-notification-activated", high_priority=high_priority)
+            message, title, action="app.download-notification-activated",
+            high_priority=high_priority
+        )
 
     def _show_private_chat_notification(self, user, message, title=None):
+
         self._show_notification(
-            message, title, action="app.private-chat-notification-activated", action_target=user, high_priority=True)
+            message, title, action="app.private-chat-notification-activated", action_target=user,
+            high_priority=True
+        )
+        self.window.set_urgency_hint(True)
 
     def _show_search_notification(self, search_token, message, title=None):
+
         self._show_notification(
-            message, title, action="app.search-notification-activated", action_target=search_token, high_priority=True)
+            message, title, action="app.search-notification-activated", action_target=search_token,
+            high_priority=True
+        )
 
     # Core Events #
 
@@ -513,16 +485,19 @@ class Application:
         should_finish_uploads = dialog.get_option_value()
 
         if response_id == "quit":
-            core.quit(should_finish_uploads=should_finish_uploads)
+            if should_finish_uploads:
+                core.uploads.request_shutdown()
+            else:
+                core.quit()
 
         elif response_id == "run_background":
             self.window.hide()
 
-    def on_confirm_quit(self, only_on_active_uploads=False):
+    def on_confirm_quit(self):
 
         has_active_uploads = core.uploads.has_active_uploads()
 
-        if not self.window.is_visible() or only_on_active_uploads and not has_active_uploads:
+        if not self.window.is_visible():
             # Never show confirmation dialog when main window is hidden
             core.quit()
             return
@@ -538,12 +513,10 @@ class Application:
 
         buttons = [
             ("cancel", _("_No")),
-            ("quit", _("_Quit"))
+            ("quit", _("_Quit")),
+            ("run_background", _("_Run in Background"))
         ]
 
-        if not only_on_active_uploads:
-            buttons.append(("run_background", _("_Run in Background")))
-
         OptionDialog(
             parent=self.window,
             title=_("Quit Nicotine+"),
@@ -579,27 +552,8 @@ class Application:
             callback=self.on_shares_unavailable_response
         ).present()
 
-    def on_invalid_password_response(self, *_args):
-        self.on_preferences(page_id="network")
-
     def on_invalid_password(self):
-
-        from pynicotine.gtkgui.widgets.dialogs import OptionDialog
-
-        title = _("Invalid Password")
-        msg = _("User %s already exists, and the password you entered is invalid. Please choose another username "
-                "if this is your first time logging in.") % config.sections["server"]["login"]
-
-        OptionDialog(
-            parent=self.window,
-            title=title,
-            message=msg,
-            buttons=[
-                ("cancel", _("_Cancel")),
-                ("ok", _("Change _Login Details"))
-            ],
-            callback=self.on_invalid_password_response
-        ).present()
+        self.on_fast_configure(invalid_password=True)
 
     def on_user_status(self, msg):
         if msg.user == core.users.login_username:
@@ -661,12 +615,16 @@ class Application:
     def on_debug_miscellaneous(self, action, state):
         self.on_set_debug_level(action, state, "miscellaneous")
 
-    def on_fast_configure(self, *_args):
+    def on_fast_configure(self, *_args, invalid_password=False):
 
         if self.fast_configure is None:
             from pynicotine.gtkgui.dialogs.fastconfigure import FastConfigure
             self.fast_configure = FastConfigure(self)
 
+        if invalid_password and self.fast_configure.is_visible():
+            self.fast_configure.hide()
+
+        self.fast_configure.invalid_password = invalid_password
         self.fast_configure.present()
 
     def on_keyboard_shortcuts(self, *_args):
@@ -848,14 +806,6 @@ class Application:
 
         self.window.present()
 
-    def on_connect_disconnect(self, *_args):
-
-        if core.users.login_status != slskmessages.UserStatus.OFFLINE:
-            self.on_disconnect()
-            return
-
-        self.on_connect()
-
     def on_away_accelerator(self, action, *_args):
         """Ctrl+H: Away/Online toggle."""
 
@@ -869,17 +819,14 @@ class Application:
     def on_away(self, *_args):
         """Away/Online status button."""
 
+        if core.users.login_status == UserStatus.OFFLINE:
+            core.connect()
+            return
+
         core.users.set_away_mode(core.users.login_status != UserStatus.AWAY, save_state=True)
 
     # Running #
 
-    def _force_quit(self):
-        """Used when the thread event processor fails due to an unhandled
-        exception, to force a shutdown."""
-
-        core.quit()
-        events.emit("quit")
-
     def _raise_exception(self, exc_value):
         raise exc_value
 
@@ -919,7 +866,7 @@ class Application:
     def _on_critical_error(self, exc_type, exc_value, exc_traceback):
 
         if self.ci_mode:
-            self._force_quit()
+            core.quit()
             self._raise_exception(exc_value)
             return
 
@@ -955,7 +902,12 @@ class Application:
 
         # Dialog was closed, quit
         sys.excepthook = None
-        self._force_quit()
+        core.quit()
+
+        # Process 'quit' event after slight delay in case thread event loop is stuck
+        GLib.idle_add(lambda: events.process_thread_events() == -1)
+
+        # Log exception in terminal
         self._raise_exception(exc_value)
 
     def on_critical_error(self, _exc_type, exc_value, _exc_traceback):
@@ -1010,11 +962,16 @@ class Application:
     def on_confirm_quit_request(self, *_args):
         core.confirm_quit()
 
-    def on_confirm_quit_uploads_request(self, *_args):
-        core.confirm_quit(only_on_active_uploads=True)
+    def on_force_quit_request(self, *_args):
+        core.quit()
 
     def on_quit_request(self, *_args):
-        core.quit()
+
+        if not core.uploads.has_active_uploads():
+            core.quit()
+            return
+
+        core.confirm_quit()
 
     def on_shutdown(self, *_args):
 
