@@ -135,16 +135,13 @@ class Buddies:
         )
 
         # Popup menus
-        self.popup_menu_private_rooms = UserPopupMenu(window.application)
-
         self.popup_menu = popup = UserPopupMenu(
             window.application, parent=self.list_view.widget, callback=self.on_popup_menu,
             tab_name="userlist"
         )
         popup.add_items(
-            ("", None),
             ("#" + _("Add User _Note…"), self.on_add_note),
-            (">" + _("Private Rooms"), self.popup_menu_private_rooms),
+            ("", None),
             ("#" + _("Remove"), self.on_remove_buddy)
         )
 
@@ -158,6 +155,7 @@ class Buddies:
             ("buddy-trusted", self.buddy_trusted),
             ("remove-buddy", self.remove_buddy),
             ("server-disconnect", self.server_disconnect),
+            ("start", self.start),
             ("user-country", self.user_country),
             ("user-stats", self.user_stats),
             ("user-status", self.user_status)
@@ -166,11 +164,31 @@ class Buddies:
 
         self.set_buddy_list_position()
 
+    def start(self):
+
+        comboboxes = (
+            self.window.search.user_search_combobox,
+            self.window.userbrowse.userbrowse_combobox,
+            self.window.userinfo.userinfo_combobox
+        )
+
+        for combobox in comboboxes:
+            combobox.freeze()
+
+        self.list_view.freeze()
+
+        for username, user_data in core.buddies.users.items():
+            self.add_buddy(username, user_data, select_row=False)
+
+        for combobox in comboboxes:
+            combobox.unfreeze()
+
+        self.list_view.unfreeze()
+
     def destroy(self):
 
         self.list_view.destroy()
         self.popup_menu.destroy()
-        self.popup_menu_private_rooms.destroy()
 
         self.__dict__.clear()
 
@@ -178,7 +196,7 @@ class Buddies:
 
         self.update_visible()
 
-        if self.container.get_visible():
+        if self.container.get_parent().get_visible():
             self.list_view.grab_focus()
             return True
 
@@ -272,10 +290,6 @@ class Buddies:
         username = self.get_selected_username()
         menu.set_user(username)
         menu.toggle_user_items()
-        menu.populate_private_rooms(self.popup_menu_private_rooms)
-
-        private_rooms_enabled = (self.popup_menu_private_rooms.items and username != core.users.login_username)
-        menu.actions[_("Private Rooms")].set_enabled(private_rooms_enabled)
 
     def user_status(self, msg):
 
@@ -287,31 +301,32 @@ class Buddies:
         status = msg.status
         status_icon_name = USER_STATUS_ICON_NAMES.get(status)
 
-        if not status_icon_name or status_icon_name == self.list_view.get_row_value(iterator, "status"):
-            return
-
-        self.list_view.set_row_value(iterator, "status", status_icon_name)
+        if status_icon_name and status_icon_name != self.list_view.get_row_value(iterator, "status"):
+            self.list_view.set_row_value(iterator, "status", status_icon_name)
 
     def user_stats(self, msg):
 
-        user = msg.user
-        iterator = self.list_view.iterators.get(user)
+        iterator = self.list_view.iterators.get(msg.user)
 
         if iterator is None:
             return
 
-        speed = msg.avgspeed
-        files = msg.files
+        speed = msg.avgspeed or 0
+        num_files = msg.files or 0
 
-        h_speed = human_speed(speed) if speed > 0 else ""
-        h_files = humanize(files)
+        if speed != self.list_view.get_row_value(iterator, "speed_data"):
+            h_speed = human_speed(speed) if speed > 0 else ""
 
-        self.list_view.set_row_value(iterator, "speed", h_speed)
-        self.list_view.set_row_value(iterator, "files", h_files)
-        self.list_view.set_row_value(iterator, "speed_data", speed)
-        self.list_view.set_row_value(iterator, "files_data", files)
+            self.list_view.set_row_value(iterator, "speed", h_speed)
+            self.list_view.set_row_value(iterator, "speed_data", speed)
 
-    def add_buddy(self, user, user_data):
+        if num_files != self.list_view.get_row_value(iterator, "files_data"):
+            h_num_files = humanize(num_files)
+
+            self.list_view.set_row_value(iterator, "files", h_num_files)
+            self.list_view.set_row_value(iterator, "files_data", num_files)
+
+    def add_buddy(self, user, user_data, select_row=True):
 
         status = user_data.status
         country_code = user_data.country.replace("flag_", "")
@@ -353,16 +368,14 @@ class Buddies:
             speed,
             files or 0,
             last_seen
-        ], select_row=core.buddies.allow_saving_buddies)
+        ], select_row=select_row)
 
         for combobox in (
             self.window.search.user_search_combobox,
             self.window.userbrowse.userbrowse_combobox,
             self.window.userinfo.userinfo_combobox
         ):
-            combobox.freeze()
             combobox.append(str(user))
-            combobox.unfreeze()
 
         self.update_visible()
 
@@ -381,9 +394,7 @@ class Buddies:
             self.window.userbrowse.userbrowse_combobox,
             self.window.userinfo.userinfo_combobox
         ):
-            combobox.freeze()
             combobox.remove_id(user)
-            combobox.unfreeze()
 
     def buddy_note(self, user, note):
 
@@ -439,10 +450,8 @@ class Buddies:
 
         flag_icon_name = get_flag_icon_name(country_code)
 
-        if not flag_icon_name:
-            return
-
-        self.list_view.set_row_value(iterator, "country", flag_icon_name)
+        if flag_icon_name and flag_icon_name != self.list_view.get_row_value(iterator, "country"):
+            self.list_view.set_row_value(iterator, "country", flag_icon_name)
 
     def on_add_buddy(self, *_args):
 
@@ -514,10 +523,5 @@ class Buddies:
         ).present()
 
     def server_disconnect(self, *_args):
-
         for iterator in self.list_view.iterators.values():
             self.list_view.set_row_value(iterator, "status", USER_STATUS_ICON_NAMES[UserStatus.OFFLINE])
-            self.list_view.set_row_value(iterator, "speed", "")
-            self.list_view.set_row_value(iterator, "files", "")
-            self.list_view.set_row_value(iterator, "speed_data", 0)
-            self.list_view.set_row_value(iterator, "files_data", 0)
