@@ -113,7 +113,7 @@ class Interests:
                 },
 
                 # Hidden data columns
-                "rating_data": {"data_type": int}
+                "rating_data": {"data_type": GObject.TYPE_INT}
             }
         )
 
@@ -165,13 +165,19 @@ class Interests:
             }
         )
 
+        self.likes_list_view.freeze()
+        self.dislikes_list_view.freeze()
+
         for item in config.sections["interests"]["likes"]:
             if isinstance(item, str):
-                self.add_thing_i_like(item)
+                self.add_thing_i_like(item, select_row=False)
 
         for item in config.sections["interests"]["dislikes"]:
             if isinstance(item, str):
-                self.add_thing_i_hate(item)
+                self.add_thing_i_hate(item, select_row=False)
+
+        self.likes_list_view.unfreeze()
+        self.dislikes_list_view.unfreeze()
 
         # Popup menus
         popup = PopupMenu(self.window.application, self.likes_list_view.widget)
@@ -273,11 +279,35 @@ class Interests:
         if self.populated_recommends or core.users.login_status == UserStatus.OFFLINE:
             return
 
-        self.on_recommendations_clicked()
+        self.show_recommendations()
 
+    def show_recommendations(self):
+
+        self.recommendations_label.set_label(_("Recommendations"))
+        self.similar_users_label.set_label(_("Similar Users"))
+
+        if not self.likes_list_view.iterators and not self.dislikes_list_view.iterators:
+            core.interests.request_global_recommendations()
+        else:
+            core.interests.request_recommendations()
+
+        core.interests.request_similar_users()
         self.populated_recommends = True
 
-    def add_thing_i_like(self, item):
+    def show_item_recommendations(self, list_view, column_id):
+
+        for iterator in list_view.get_selected_rows():
+            item = list_view.get_row_value(iterator, column_id)
+
+            core.interests.request_item_recommendations(item)
+            core.interests.request_item_similar_users(item)
+            self.populated_recommends = True
+
+            if self.window.current_page_id != self.window.interests_page.id:
+                self.window.change_main_page(self.window.interests_page)
+            return
+
+    def add_thing_i_like(self, item, select_row=True):
 
         item = item.strip().lower()
 
@@ -287,9 +317,9 @@ class Interests:
         iterator = self.likes_list_view.iterators.get(item)
 
         if iterator is None:
-            self.likes_list_view.add_row([item])
+            self.likes_list_view.add_row([item], select_row=select_row)
 
-    def add_thing_i_hate(self, item):
+    def add_thing_i_hate(self, item, select_row=True):
 
         item = item.strip().lower()
 
@@ -299,7 +329,7 @@ class Interests:
         iterator = self.dislikes_list_view.iterators.get(item)
 
         if iterator is None:
-            self.dislikes_list_view.add_row([item])
+            self.dislikes_list_view.add_row([item], select_row=select_row)
 
     def remove_thing_i_like(self, item):
 
@@ -315,9 +345,6 @@ class Interests:
         if iterator is not None:
             self.dislikes_list_view.remove_row(iterator)
 
-    def recommend_search(self, item):
-        core.search.do_search(item, mode="global")
-
     def on_add_thing_i_like(self, *_args):
 
         item = self.add_like_entry.get_text().strip()
@@ -381,33 +408,18 @@ class Interests:
             return
 
     def on_recommend_item(self, _action, _state, list_view, column_id):
-
-        for iterator in list_view.get_selected_rows():
-            item = list_view.get_row_value(iterator, column_id)
-
-            core.interests.request_item_recommendations(item)
-            core.interests.request_item_similar_users(item)
-            return
+        self.show_item_recommendations(list_view, column_id)
 
     def on_recommend_search(self, _action, _state, list_view, column_id):
 
         for iterator in list_view.get_selected_rows():
             item = list_view.get_row_value(iterator, column_id)
 
-            self.recommend_search(item)
+            core.search.do_search(item, mode="global")
             return
 
-    def on_recommendations_clicked(self, *_args):
-
-        self.recommendations_label.set_label(_("Recommendations"))
-        self.similar_users_label.set_label(_("Similar Users"))
-
-        if not self.likes_list_view.iterators and not self.dislikes_list_view.iterators:
-            core.interests.request_global_recommendations()
-        else:
-            core.interests.request_recommendations()
-
-        core.interests.request_similar_users()
+    def on_refresh_recommendations(self, *_args):
+        self.show_recommendations()
 
     def set_recommendations(self, recommendations, item=None):
 
@@ -417,10 +429,13 @@ class Interests:
             self.recommendations_label.set_label(_("Recommendations"))
 
         self.recommendations_list_view.clear()
+        self.recommendations_list_view.freeze()
 
         for thing, rating in recommendations:
             self.recommendations_list_view.add_row([humanize(rating), thing, rating], select_row=False)
 
+        self.recommendations_list_view.unfreeze()
+
     def global_recommendations(self, msg):
         self.set_recommendations(msg.recommendations + msg.unrecommendations)
 
@@ -438,8 +453,9 @@ class Interests:
             self.similar_users_label.set_label(_("Similar Users"))
 
         self.similar_users_list_view.clear()
+        self.similar_users_list_view.freeze()
 
-        for index, (user, rating) in enumerate(users.items()):
+        for index, (user, rating) in enumerate(reversed(list(users.items()))):
             status = core.users.statuses.get(user, UserStatus.OFFLINE)
             country_code = core.users.countries.get(user, "")
             stats = core.users.watched.get(user)
@@ -466,6 +482,8 @@ class Interests:
                 rating
             ], select_row=False)
 
+        self.similar_users_list_view.unfreeze()
+
     def similar_users(self, msg):
         self.set_similar_users(msg.users)
 
@@ -482,10 +500,8 @@ class Interests:
 
         flag_icon_name = get_flag_icon_name(country_code)
 
-        if not flag_icon_name:
-            return
-
-        self.similar_users_list_view.set_row_value(iterator, "country", flag_icon_name)
+        if flag_icon_name and flag_icon_name != self.similar_users_list_view.get_row_value(iterator, "country"):
+            self.similar_users_list_view.set_row_value(iterator, "country", flag_icon_name)
 
     def user_status(self, msg):
 
@@ -496,10 +512,8 @@ class Interests:
 
         status_icon_name = USER_STATUS_ICON_NAMES.get(msg.status)
 
-        if not status_icon_name or status_icon_name == self.similar_users_list_view.get_row_value(iterator, "status"):
-            return
-
-        self.similar_users_list_view.set_row_value(iterator, "status", status_icon_name)
+        if status_icon_name and status_icon_name != self.similar_users_list_view.get_row_value(iterator, "status"):
+            self.similar_users_list_view.set_row_value(iterator, "status", status_icon_name)
 
     def user_stats(self, msg):
 
@@ -508,16 +522,20 @@ class Interests:
         if iterator is None:
             return
 
-        speed = msg.avgspeed
-        files = msg.files
+        speed = msg.avgspeed or 0
+        num_files = msg.files or 0
 
-        h_speed = human_speed(speed) if speed > 0 else ""
-        h_files = humanize(msg.files)
+        if speed != self.similar_users_list_view.get_row_value(iterator, "speed_data"):
+            h_speed = human_speed(speed) if speed > 0 else ""
 
-        self.similar_users_list_view.set_row_value(iterator, "speed", h_speed)
-        self.similar_users_list_view.set_row_value(iterator, "files", h_files)
-        self.similar_users_list_view.set_row_value(iterator, "speed_data", speed)
-        self.similar_users_list_view.set_row_value(iterator, "files_data", files)
+            self.similar_users_list_view.set_row_value(iterator, "speed", h_speed)
+            self.similar_users_list_view.set_row_value(iterator, "speed_data", speed)
+
+        if num_files != self.similar_users_list_view.get_row_value(iterator, "files_data"):
+            h_num_files = humanize(num_files)
+
+            self.similar_users_list_view.set_row_value(iterator, "files", h_num_files)
+            self.similar_users_list_view.set_row_value(iterator, "files_data", num_files)
 
     @staticmethod
     def toggle_menu_items(menu, list_view, column_id):
@@ -537,13 +555,7 @@ class Interests:
         self.toggle_menu_items(menu, self.recommendations_list_view, column_id="item")
 
     def on_r_row_activated(self, *_args):
-
-        for iterator in self.recommendations_list_view.get_selected_rows():
-            item = self.recommendations_list_view.get_row_value(iterator, "item")
-
-            core.interests.request_item_recommendations(item)
-            core.interests.request_item_similar_users(item)
-            return
+        self.show_item_recommendations(self.recommendations_list_view, column_id="item")
 
     def on_popup_ru_menu(self, menu, *_args):
 
