@@ -22,6 +22,7 @@
 # You should have received a copy of the GNU General Public License
 # along with this program.  If not, see <http://www.gnu.org/licenses/>.
 
+import os
 import sys
 import time
 
@@ -57,7 +58,6 @@ from pynicotine.gtkgui.widgets.theme import set_use_header_bar
 from pynicotine.gtkgui.widgets.window import Window
 from pynicotine.logfacility import log
 from pynicotine.slskmessages import UserStatus
-from pynicotine.utils import human_speed
 from pynicotine.utils import open_folder_path
 
 
@@ -71,6 +71,8 @@ class MainWindow(Window):
         self.away_timer_id = None
         self.away_cooldown_time = 0
         self.gesture_click = None
+        self.window_active_handler = None
+        self.window_visible_handler = None
 
         # Load UI
 
@@ -107,6 +109,7 @@ class MainWindow(Window):
             self.header_end_container,
             self.header_menu,
             self.header_title,
+            self.hide_window_button,
             self.horizontal_paned,
             self.interests_container,
             self.interests_end,
@@ -115,7 +118,6 @@ class MainWindow(Window):
             self.interests_toolbar,
             self.log_container,
             self.log_search_bar,
-            self.log_search_entry,
             self.log_view_container,
             self.private_content,
             self.private_end,
@@ -219,25 +221,26 @@ class MainWindow(Window):
             self.vertical_paned.child_set_property(self.log_container, "shrink", False)
 
         # Logging
-        self.log_view = TextView(self.log_view_container, auto_scroll=not config.sections["logging"]["logcollapsed"],
-                                 parse_urls=False, editable=False, vertical_margin=5, pixels_below_lines=2)
-        self.log_search_bar = TextSearchBar(self.log_view.widget, self.log_search_bar, self.log_search_entry,
-                                            controller_widget=self.log_container)
+        self.log_view = TextView(
+            self.log_view_container, auto_scroll=not config.sections["logging"]["logcollapsed"],
+            parse_urls=False, editable=False, vertical_margin=5, pixels_below_lines=2
+        )
+        self.log_search_bar = TextSearchBar(
+            self.log_view.widget, self.log_search_bar, controller_widget=self.log_container,
+            placeholder_text=_("Search log…")
+        )
 
         self.create_log_context_menu()
         events.connect("log-message", self.log_callback)
 
         # Events
         for event_name, callback in (
-            ("schedule-quit", self.schedule_quit),
             ("server-login", self.update_user_status),
             ("server-disconnect", self.update_user_status),
             ("set-connection-stats", self.set_connection_stats),
             ("shares-preparing", self.shares_preparing),
             ("shares-ready", self.shares_ready),
             ("shares-scanning", self.shares_scanning),
-            ("update-download-limits", self.update_download_limits),
-            ("update-upload-limits", self.update_upload_limits),
             ("user-status", self.user_status)
         ):
             events.connect(event_name, callback)
@@ -295,6 +298,8 @@ class MainWindow(Window):
 
     def init_window(self):
 
+        is_broadway_backend = (os.environ.get("GDK_BACKEND") == "broadway")
+
         # Set main window title and icon
         self.set_title(pynicotine.__application_name__)
         self.widget.set_default_icon_name(pynicotine.__application_id__)
@@ -313,8 +318,12 @@ class MainWindow(Window):
             else:
                 self.widget.move(x_pos, y_pos)
 
+        # Hide close button in Broadway backend
+        if is_broadway_backend:
+            self.widget.set_deletable(False)
+
         # Maximize main window if necessary
-        if sys.platform != "darwin" and config.sections["ui"]["maximized"]:
+        if config.sections["ui"]["maximized"] or is_broadway_backend:
             self.widget.maximize()
 
         # Auto-away mode
@@ -336,8 +345,8 @@ class MainWindow(Window):
         self.gesture_click.connect("pressed", self.on_cancel_auto_away)
 
         # Clear notifications when main window is focused
-        self.widget.connect("notify::is-active", self.on_window_active_changed)
-        self.widget.connect("notify::visible", self.on_window_visible_changed)
+        self.window_active_handler = self.widget.connect("notify::is-active", self.on_window_active_changed)
+        self.window_visible_handler = self.widget.connect("notify::visible", self.on_window_visible_changed)
 
         # System window close (X)
         if GTK_API_VERSION >= 4:
@@ -366,7 +375,7 @@ class MainWindow(Window):
         self.set_urgency_hint(False)
 
     def on_window_visible_changed(self, *_args):
-        self.application.tray_icon.update_window_visibility()
+        self.application.tray_icon.update()
 
     def update_title(self):
 
@@ -380,17 +389,17 @@ class MainWindow(Window):
             # Private Chats have a higher priority
             user = self.privatechat.highlighted_users[-1]
             notification_text = _("Private Message from %(user)s") % {"user": user}
+            self.set_urgency_hint(True)
 
         elif self.chatrooms.highlighted_rooms:
             # Allow for the possibility the username is not available
             room, user = list(self.chatrooms.highlighted_rooms.items())[-1]
             notification_text = _("Mentioned by %(user)s in Room %(room)s") % {"user": user, "room": room}
+            self.set_urgency_hint(True)
 
         elif any(is_important for is_important in self.search.unread_pages.values()):
             notification_text = _("Wishlist Results Found")
 
-        self.set_urgency_hint(bool(notification_text))
-
         if not notification_text:
             self.set_title(pynicotine.__application_name__)
             return
@@ -457,7 +466,6 @@ class MainWindow(Window):
         self.add_action(action)
 
         action = Gio.SimpleAction(name="toggle-status")
-        action.set_enabled(False)
         action.connect("activate", self.on_toggle_status)
         self.add_action(action)
 
@@ -540,6 +548,15 @@ class MainWindow(Window):
         end_widget = self.tabs[page_id].toolbar_end_content
         end_widget.get_parent().remove(end_widget)
 
+        for widget in end_widget:
+            # Themes decide if header bar buttons should be flat
+            if isinstance(widget, Gtk.Button):
+                remove_css_class(widget, "flat")
+
+            # Header bars never contain separators, hide them
+            elif isinstance(widget, Gtk.Separator):
+                widget.set_visible(False)
+
         if GTK_API_VERSION >= 4:
             self.header_title.append(title_widget)
             self.header_end_container.append(end_widget)
@@ -588,6 +605,15 @@ class MainWindow(Window):
                 self.widget.set_titlebar(None)
                 self.widget.map()
 
+        for widget in self.tabs[page_id].toolbar_end_content:
+            # Make secondary buttons at the end of the toolbar flat. Keep buttons
+            # next to text entries raised for more prominence.
+            if isinstance(widget, Gtk.Button):
+                add_css_class(widget, "flat")
+
+            elif isinstance(widget, Gtk.Separator):
+                widget.set_visible(True)
+
         toolbar = self.tabs[page_id].toolbar
         toolbar.set_visible(True)
 
@@ -900,12 +926,7 @@ class MainWindow(Window):
     def update_user_status(self, *_args):
 
         status = core.users.login_status
-        is_online = (status != UserStatus.OFFLINE)
         is_away = (status == UserStatus.AWAY)
-        toggle_status_action = self.lookup_action("toggle-status")
-
-        # Action status
-        toggle_status_action.set_enabled(is_online)
 
         # Away mode
         if not is_away:
@@ -914,9 +935,6 @@ class MainWindow(Window):
             self.remove_away_timer()
 
         # Status bar
-        if core.uploads.pending_shutdown:
-            return
-
         username = core.users.login_username
         icon_name = USER_STATUS_ICON_NAMES[status]
         icon_args = (Gtk.IconSize.BUTTON,) if GTK_API_VERSION == 3 else ()  # pylint: disable=no-member
@@ -942,10 +960,11 @@ class MainWindow(Window):
             self.user_status_label.set_text(status_text)
 
         if self.user_status_button.get_active():
-            # Disable button toggled state without activating action
-            toggle_status_action.handler_block_by_func(self.on_toggle_status)
+            toggle_status_action = self.lookup_action("toggle-status")
+
+            toggle_status_action.set_enabled(False)
             self.user_status_button.set_active(False)
-            toggle_status_action.handler_unblock_by_func(self.on_toggle_status)
+            toggle_status_action.set_enabled(True)
 
     def user_status(self, msg):
         if msg.user == core.users.login_username:
@@ -956,6 +975,17 @@ class MainWindow(Window):
     def on_search(self, *_args):
         self.search.on_search()
 
+    def on_search_entry_changed(self, entry, *_args):
+        entry.props.secondary_icon_name = "edit-clear-symbolic" if entry.get_text() else None
+
+    def on_search_entry_icon_press(self, entry, icon_pos, *_args):
+
+        if icon_pos == Gtk.EntryIconPosition.SECONDARY:
+            entry.set_text("")
+            return
+
+        self.on_search()
+
     # User Info #
 
     def on_show_user_profile(self, *_args):
@@ -1027,8 +1057,8 @@ class MainWindow(Window):
 
     def create_log_context_menu(self):
 
-        popup_menu_log_categories = PopupMenu(self.application)
-        popup_menu_log_categories.add_items(
+        self.popup_menu_log_categories = PopupMenu(self.application)
+        self.popup_menu_log_categories.add_items(
             ("$" + _("Downloads"), "app.log-downloads"),
             ("$" + _("Uploads"), "app.log-uploads"),
             ("$" + _("Search"), "app.log-searches"),
@@ -1040,7 +1070,8 @@ class MainWindow(Window):
             ("$" + _("[Debug] Miscellaneous"), "app.log-miscellaneous"),
         )
 
-        PopupMenu(self.application, self.log_view.widget, self.on_popup_menu_log).add_items(
+        self.popup_menu_log_view = PopupMenu(self.application, self.log_view.widget, self.on_popup_menu_log)
+        self.popup_menu_log_view.add_items(
             ("#" + _("_Find…"), self.on_find_log_window),
             ("", None),
             ("#" + _("_Copy"), self.log_view.on_copy_text),
@@ -1049,7 +1080,7 @@ class MainWindow(Window):
             ("#" + _("View _Debug Logs"), self.on_view_debug_logs),
             ("#" + _("View _Transfer Logs"), self.on_view_transfer_logs),
             ("", None),
-            (">" + _("_Log Categories"), popup_menu_log_categories),
+            (">" + _("_Log Categories"), self.popup_menu_log_categories),
             ("", None),
             ("#" + _("Clear Log View"), self.on_clear_log_view)
         )
@@ -1107,48 +1138,13 @@ class MainWindow(Window):
         self.status_label.set_tooltip_text(msg)
         self.status_label.set_visible(True)
 
-    def set_connection_stats(self, total_conns=0, download_bandwidth=0, upload_bandwidth=0):
+    def set_connection_stats(self, total_conns=0, **_kwargs):
 
         total_conns_text = repr(total_conns)
-        download_bandwidth = human_speed(download_bandwidth)
-        upload_bandwidth = human_speed(upload_bandwidth)
-        download_bandwidth_text = f"{download_bandwidth} ( {len(core.downloads.active_users)} )"
-        upload_bandwidth_text = f"{upload_bandwidth} ( {len(core.uploads.active_users)} )"
 
         if self.connections_label.get_text() != total_conns_text:
             self.connections_label.set_text(total_conns_text)
 
-        if self.download_status_label.get_text() != download_bandwidth_text:
-            self.download_status_label.set_text(download_bandwidth_text)
-            self.application.tray_icon.set_download_status(_("Downloads: %(speed)s") % {"speed": download_bandwidth})
-
-        if self.upload_status_label.get_text() != upload_bandwidth_text:
-            self.upload_status_label.set_text(upload_bandwidth_text)
-            self.application.tray_icon.set_upload_status(_("Uploads: %(speed)s") % {"speed": upload_bandwidth})
-
-    def update_download_limits(self):
-        self.update_bandwidth_label_underlines(transfer_type="download")
-
-    def update_upload_limits(self):
-        self.update_bandwidth_label_underlines(transfer_type="upload")
-
-    def update_bandwidth_label_underlines(self, transfer_type):
-        """Underline status bar bandwidth labels when alternative speed limits
-        are active."""
-
-        if transfer_type == "download":
-            label = self.download_status_label
-            config_key = "use_download_speed_limit"
-        else:
-            label = self.upload_status_label
-            config_key = "use_upload_speed_limit"
-
-        if config.sections["transfers"][config_key] == "alternative":
-            add_css_class(label, "underline")
-            return
-
-        remove_css_class(label, "underline")
-
     def shares_preparing(self):
 
         # Hide widget to keep tooltips for other widgets visible
@@ -1166,18 +1162,17 @@ class MainWindow(Window):
         self.scan_progress_spinner.start()
 
     def shares_ready(self, _successful):
-
         self.scan_progress_container.set_visible(False)
         self.scan_progress_spinner.stop()
 
     def on_toggle_status(self, *_args):
 
         if core.uploads.pending_shutdown:
-            core.uploads.pending_shutdown = False
-            self.update_user_status()
-            return
+            core.uploads.cancel_shutdown()
+        else:
+            self.application.lookup_action("away").activate()
 
-        self.application.lookup_action("away").activate()
+        self.user_status_button.set_active(False)
 
     # Exit #
 
@@ -1194,22 +1189,6 @@ class MainWindow(Window):
 
         return True
 
-    def schedule_quit(self, should_finish_uploads):
-
-        if not should_finish_uploads:
-            return
-
-        icon_name = "system-shutdown-symbolic"
-        icon_args = (Gtk.IconSize.BUTTON,) if GTK_API_VERSION == 3 else ()  # pylint: disable=no-member
-        toggle_status_action = self.lookup_action("toggle-status")
-
-        toggle_status_action.handler_block_by_func(self.on_toggle_status)
-        self.user_status_button.set_active(True)
-        toggle_status_action.handler_unblock_by_func(self.on_toggle_status)
-
-        self.user_status_icon.set_from_icon_name(icon_name, *icon_args)
-        self.user_status_label.set_text(_("Quitting..."))
-
     def hide(self):
 
         if not self.is_visible():
@@ -1223,6 +1202,17 @@ class MainWindow(Window):
         config.write_configuration()
 
         # Hide window
+        if sys.platform == "darwin":
+            # macOS-specific way to hide the application, to ensure it is restored when clicking the dock icon
+            self.hide_window_button.set_action_name("gtkinternal.hide")
+            self.hide_window_button.emit("clicked")
+            return
+
+        if GTK_API_VERSION >= 4:
+            self.widget.minimize()
+        else:
+            self.widget.iconify()
+
         super().hide()
 
     def destroy(self):
@@ -1230,4 +1220,13 @@ class MainWindow(Window):
         for tab in self.tabs.values():
             tab.destroy()
 
+        self.notebook.destroy()
+        self.log_search_bar.destroy()
+        self.log_view.destroy()
+        self.popup_menu_log_view.destroy()
+        self.popup_menu_log_categories.destroy()
+
+        self.widget.disconnect(self.window_active_handler)
+        self.widget.disconnect(self.window_visible_handler)
+
         super().destroy()
