@@ -21,8 +21,6 @@
 # You should have received a copy of the GNU General Public License
 # along with this program.  If not, see <http://www.gnu.org/licenses/>.
 
-import os
-
 from gi.repository import GLib
 
 from pynicotine.config import config
@@ -35,14 +33,11 @@ from pynicotine.gtkgui.widgets.iconnotebook import IconNotebook
 from pynicotine.gtkgui.widgets.popupmenu import PopupMenu
 from pynicotine.gtkgui.widgets.popupmenu import UserPopupMenu
 from pynicotine.gtkgui.widgets.dialogs import OptionDialog
-from pynicotine.gtkgui.widgets.textentry import ChatCompletion
 from pynicotine.gtkgui.widgets.textentry import ChatEntry
-from pynicotine.gtkgui.widgets.textentry import SpellChecker
 from pynicotine.gtkgui.widgets.textentry import TextSearchBar
 from pynicotine.gtkgui.widgets.textview import ChatView
 from pynicotine.logfacility import log
 from pynicotine.slskmessages import UserStatus
-from pynicotine.utils import clean_file
 
 
 class PrivateChats(IconNotebook):
@@ -64,11 +59,14 @@ class PrivateChats(IconNotebook):
         self.toolbar_end_content = window.private_end
         self.toolbar_default_widget = window.private_entry
 
-        self.highlighted_users = []
-        self.completion = ChatCompletion()
-        self.spell_checker = SpellChecker()
+        self.chat_entry = ChatEntry(
+            window.application, send_message_callback=core.privatechat.send_message,
+            command_callback=core.pluginhandler.trigger_private_chat_command_event,
+            enable_spell_check=config.sections["ui"]["spellcheck"]
+        )
         self.history = ChatHistory(window)
         self.command_help = None
+        self.highlighted_users = []
 
         for event_name, callback in (
             ("clear-private-messages", self.clear_messages),
@@ -77,16 +75,25 @@ class PrivateChats(IconNotebook):
             ("private-chat-completions", self.update_completions),
             ("private-chat-show-user", self.show_user),
             ("private-chat-remove-user", self.remove_user),
+            ("quit", self.quit),
             ("server-disconnect", self.server_disconnect),
             ("server-login", self.server_login),
+            ("start", self.start),
             ("user-status", self.user_status)
         ):
             events.connect(event_name, callback)
 
+        self.freeze()
+
+    def start(self):
+        self.unfreeze()
+
+    def quit(self):
+        self.freeze()
+
     def destroy(self):
 
-        self.completion.destroy()
-        self.spell_checker.destroy()
+        self.chat_entry.destroy()
         self.history.destroy()
 
         if self.command_help is not None:
@@ -134,8 +141,7 @@ class PrivateChats(IconNotebook):
             if tab.container != page:
                 continue
 
-            self.spell_checker.set_entry(tab.chat_entry)
-            self.completion.set_entry(tab.chat_entry)
+            self.chat_entry.set_parent(user, tab.chat_entry_container, tab.chat_view)
             tab.update_room_user_completions()
 
             if self.command_help is None:
@@ -217,9 +223,6 @@ class PrivateChats(IconNotebook):
             return
 
         if page.container == self.get_current_page():
-            self.spell_checker.set_entry(None)
-            self.completion.set_entry(None)
-
             if self.command_help is not None:
                 self.command_help.set_menu_button(None)
 
@@ -228,6 +231,8 @@ class PrivateChats(IconNotebook):
         del self.pages[user]
         page.destroy()
 
+        self.chat_entry.clear_unsent_message(user)
+
     def highlight_user(self, user):
 
         if not user or user in self.highlighted_users:
@@ -235,7 +240,7 @@ class PrivateChats(IconNotebook):
 
         self.highlighted_users.append(user)
         self.window.update_title()
-        self.window.application.tray_icon.update_icon()
+        self.window.application.tray_icon.update()
 
     def unhighlight_user(self, user):
 
@@ -244,7 +249,7 @@ class PrivateChats(IconNotebook):
 
         self.highlighted_users.remove(user)
         self.window.update_title()
-        self.window.application.tray_icon.update_icon()
+        self.window.application.tray_icon.update()
 
     def echo_private_message(self, user, text, message_type):
 
@@ -271,12 +276,9 @@ class PrivateChats(IconNotebook):
 
     def update_widgets(self):
 
-        page = self.get_current_page()
+        self.chat_entry.set_spell_check_enabled(config.sections["ui"]["spellcheck"])
 
         for tab in self.pages.values():
-            if tab.container == page:
-                self.spell_checker.set_entry(tab.chat_entry)
-
             tab.toggle_chat_buttons()
             tab.update_tags()
 
@@ -286,15 +288,17 @@ class PrivateChats(IconNotebook):
             return
 
         page = self.get_current_page()
+        self.chat_entry.set_sensitive(True)
 
         for tab in self.pages.values():
-            tab.server_login()
-
             if tab.container == page:
                 tab.on_focus()
+                break
 
     def server_disconnect(self, *_args):
 
+        self.chat_entry.set_sensitive(False)
+
         for user, page in self.pages.items():
             page.server_disconnect()
             self.set_user_status(page.container, user, UserStatus.OFFLINE)
@@ -305,13 +309,12 @@ class PrivateChat:
     def __init__(self, chats, user):
 
         (
-            self.chat_entry,
+            self.chat_entry_container,
             self.chat_view_container,
             self.container,
             self.help_button,
             self.log_toggle,
             self.search_bar,
-            self.search_entry,
             self.speech_toggle
         ) = ui.load(scope=self, path="privatechat.ui")
 
@@ -322,19 +325,19 @@ class PrivateChat:
         self.loaded = False
         self.offline_message = False
 
-        self.chat_view = ChatView(self.chat_view_container, chat_entry=self.chat_entry, editable=False,
-                                  horizontal_margin=10, vertical_margin=5, pixels_below_lines=2,
-                                  username_event=self.username_event)
+        self.chat_view = ChatView(
+            self.chat_view_container, chat_entry=self.chats.chat_entry, editable=False,
+            horizontal_margin=10, vertical_margin=5, pixels_below_lines=2,
+            username_event=self.username_event
+        )
 
         # Text Search
-        self.search_bar = TextSearchBar(self.chat_view.widget, self.search_bar, self.search_entry,
-                                        controller_widget=self.container, focus_widget=self.chat_entry)
+        self.search_bar = TextSearchBar(
+            self.chat_view.widget, self.search_bar,
+            controller_widget=self.container, focus_widget=self.chats.chat_entry,
+            placeholder_text=_("Search chat log…")
+        )
 
-        # Chat Entry
-        self.chat_entry = ChatEntry(self.window.application, self.chat_entry, self.chat_view, chats.completion,
-                                    user, core.privatechat.send_message)
-
-        self.chat_entry.set_sensitive(core.users.login_status != UserStatus.OFFLINE)
         self.log_toggle.set_active(user in config.sections["logging"]["private_chats"])
         self.toggle_chat_buttons()
 
@@ -367,12 +370,12 @@ class PrivateChat:
             ("", None),
             ("#" + _("Clear Message View"), self.chat_view.on_clear_all_text),
             ("", None),
-            (">" + _("User"), self.popup_menu_user_tab),
+            (">" + _("User Actions"), self.popup_menu_user_tab),
         )
 
         self.popup_menus = (self.popup_menu, self.popup_menu_user_chat, self.popup_menu_user_tab)
 
-        self.read_private_log()
+        self.prepend_old_messages()
 
     def load(self):
         GLib.idle_add(self.read_private_log_finished)
@@ -387,23 +390,19 @@ class PrivateChat:
         self.chat_view.scroll_bottom()
         self.chat_view.auto_scroll = True
 
-    def read_private_log(self):
+    def prepend_old_messages(self):
 
-        self.chat_view.append_log_lines(
-            path=os.path.join(log.private_chat_folder_path, f"{clean_file(self.user)}.log"),
-            num_lines=config.sections["logging"]["readprivatelines"],
-            timestamp_format=config.sections["logging"]["private_timestamp"]
+        log_lines = log.read_log(
+            folder_path=log.private_chat_folder_path,
+            basename=self.user,
+            num_lines=config.sections["logging"]["readprivatelines"]
         )
 
-    def server_login(self):
-        self.chat_entry.set_sensitive(True)
+        self.chat_view.append_log_lines(log_lines, login_username=config.sections["server"]["login"])
 
     def server_disconnect(self):
-
         self.offline_message = False
-
         self.chat_view.update_user_tags()
-        self.chat_entry.set_sensitive(False)
 
     def clear(self):
         self.chat_view.clear()
@@ -414,7 +413,6 @@ class PrivateChat:
         for menu in self.popup_menus:
             menu.destroy()
 
-        self.chat_entry.destroy()
         self.chat_view.destroy()
         self.search_bar.destroy()
         self.__dict__.clear()
@@ -550,7 +548,7 @@ class PrivateChat:
     def on_focus(self, *_args):
 
         if self.window.current_page_id == self.window.private_page.id:
-            widget = self.chat_entry if self.chat_entry.get_sensitive() else self.chat_view
+            widget = self.chats.chat_entry if self.chats.chat_entry.get_sensitive() else self.chat_view
             widget.grab_focus()
 
         return True
@@ -569,4 +567,4 @@ class PrivateChat:
         # Tab-complete the recipient username
         completions.add(self.user)
 
-        self.chats.completion.set_completions(completions)
+        self.chats.chat_entry.set_completions(completions)
