@@ -38,12 +38,12 @@ from pynicotine.gtkgui.dialogs.fileproperties import FileProperties
 from pynicotine.gtkgui.widgets import clipboard
 from pynicotine.gtkgui.widgets import ui
 from pynicotine.gtkgui.widgets.accelerator import Accelerator
+from pynicotine.gtkgui.widgets.combobox import ComboBox
 from pynicotine.gtkgui.widgets.filechooser import FolderChooser
 from pynicotine.gtkgui.widgets.iconnotebook import IconNotebook
 from pynicotine.gtkgui.widgets.popupmenu import PopupMenu
 from pynicotine.gtkgui.widgets.popupmenu import FilePopupMenu
 from pynicotine.gtkgui.widgets.popupmenu import UserPopupMenu
-from pynicotine.gtkgui.widgets.textentry import ComboBox
 from pynicotine.gtkgui.widgets.theme import add_css_class
 from pynicotine.gtkgui.widgets.theme import get_file_type_icon_name
 from pynicotine.gtkgui.widgets.theme import get_flag_icon_name
@@ -59,6 +59,14 @@ from pynicotine.utils import human_size
 from pynicotine.utils import human_speed
 
 
+class SearchResultFile:
+    __slots__ = ("path", "attributes")
+
+    def __init__(self, path, attributes=None):
+        self.path = path
+        self.attributes = attributes
+
+
 class Searches(IconNotebook):
 
     def __init__(self, window):
@@ -101,12 +109,8 @@ class Searches(IconNotebook):
 
         self.room_search_combobox = ComboBox(
             container=self.window.search_title, has_entry=True, has_entry_completion=True,
-            entry=self.window.room_search_entry, visible=False,
-            items=(
-                (core.chatrooms.JOINED_ROOMS_NAME, None),
-            )
+            entry=self.window.room_search_entry, visible=False
         )
-        self.room_search_combobox.set_text(core.chatrooms.JOINED_ROOMS_NAME)
 
         self.user_search_combobox = ComboBox(
             container=self.window.search_title, has_entry=True, has_entry_completion=True,
@@ -124,15 +128,20 @@ class Searches(IconNotebook):
             ("add-search", self.add_search),
             ("add-wish", self.update_wish_button),
             ("file-search-response", self.file_search_response),
+            ("quit", self.quit),
             ("remove-search", self.remove_search),
             ("remove-wish", self.update_wish_button),
-            ("server-login", self.on_focus),
+            ("server-disconnect", self.server_disconnect),
+            ("server-login", self.server_login),
             ("show-search", self.show_search)
         ):
             events.connect(event_name, callback)
 
         self.populate_search_history()
 
+    def quit(self):
+        self.freeze()
+
     def destroy(self):
 
         self.room_search_combobox.destroy()
@@ -219,16 +228,12 @@ class Searches(IconNotebook):
         if not config.sections["searches"]["enable_history"]:
             return
 
-        self.search_combobox.freeze()
-
         self.search_combobox.remove_id(term)
         self.search_combobox.prepend(term)
 
         while self.search_combobox.get_num_items() > core.search.SEARCH_HISTORY_LIMIT:
             self.search_combobox.remove_pos(-1)
 
-        self.search_combobox.unfreeze()
-
     def create_page(self, token, text, mode=None, mode_label=None, room=None, users=None,
                     show_page=True):
 
@@ -245,15 +250,10 @@ class Searches(IconNotebook):
             return page
 
         if mode_label is not None:
-            full_text = f"({mode_label}) {text}"
-            length = 25
-        else:
-            full_text = text
-            length = 20
+            text = f"({mode_label}) {text}"
 
-        label = full_text[:length]
-        self.append_page(page.container, label, focus_callback=page.on_focus,
-                         close_callback=page.on_close, full_text=full_text)
+        self.append_page(page.container, text, focus_callback=page.on_focus,
+                         close_callback=page.on_close)
         page.set_label(self.get_tab_label_inner(page.container))
 
         return page
@@ -274,12 +274,12 @@ class Searches(IconNotebook):
         elif mode == "buddies":
             mode_label = _("Buddies")
 
-        self.create_page(token, search.term, mode, mode_label, room=room, users=users)
+        self.create_page(token, search.term_sanitized, mode, mode_label, room=room, users=users)
 
         if switch_page:
             self.show_search(token)
 
-        self.add_search_history_item(search.term)
+        self.add_search_history_item(search.term_sanitized)
 
     def show_search(self, token):
 
@@ -370,6 +370,13 @@ class Searches(IconNotebook):
             if page.text == wish:
                 page.update_wish_button()
 
+    def server_login(self, *_args):
+        self.window.search_title.set_sensitive(True)
+        self.on_focus()
+
+    def server_disconnect(self, *_args):
+        self.window.search_title.set_sensitive(False)
+
 
 class Search:
 
@@ -446,6 +453,7 @@ class Search:
         self.room = room
         self.searched_users = users
         self.show_page = show_page
+        self.initialized = False
         self.users = {}
         self.folders = {}
         self.all_data = []
@@ -576,11 +584,10 @@ class Search:
                 "size_data": {"data_type": GObject.TYPE_UINT64},
                 "bitrate_data": {"data_type": GObject.TYPE_UINT},
                 "length_data": {"data_type": GObject.TYPE_UINT},
-                "file_path_data": {"data_type": str},
-                "free_slot_data": {"data_type": bool},
-                "file_attributes_data": {"data_type": GObject.TYPE_PYOBJECT},
+                "free_slot_data": {"data_type": GObject.TYPE_BOOLEAN},
+                "file_data": {"data_type": GObject.TYPE_PYOBJECT},
                 "id_data": {
-                    "data_type": GObject.TYPE_UINT64,
+                    "data_type": GObject.TYPE_INT,
                     "default_sort_type": "ascending",
                     "iterator_key": True
                 }
@@ -684,15 +691,12 @@ class Search:
             combobox.destroy()
 
         self.tree_view.destroy()
+        self.window.update_title()
         self.__dict__.clear()
 
     def set_label(self, label):
         self.tab_menu.set_parent(label)
 
-    def on_combobox_check_separator(self, model, iterator):
-        # Render empty value as separator
-        return not model.get_value(iterator, 0)
-
     def update_filter_widgets(self):
 
         self.update_filter_counter(self.active_filter_count)
@@ -715,8 +719,6 @@ class Search:
 
         for filter_id, widget in self.filter_comboboxes.items():
             widget.freeze()
-
-            widget.set_row_separator_func(lambda *_args: 0)
             widget.clear()
 
             presets = self.FILTER_PRESETS.get(filter_id)
@@ -732,9 +734,6 @@ class Search:
             for value in islice(filter_history, core.search.RESULT_FILTER_HISTORY_LIMIT):
                 widget.append(value)
 
-            if presets:
-                widget.set_row_separator_func(self.on_combobox_check_separator)
-
             widget.unfreeze()
 
     def populate_default_filters(self):
@@ -796,12 +795,8 @@ class Search:
 
             if any(word in file_path_lower for word in search.excluded_words):
                 # Filter out results with filtered words (e.g. nicotine -music)
-                log.add_debug(("Filtered out excluded search result %(filepath)s from user %(user)s for "
-                               'search term "%(query)s"'), {
-                    "filepath": file_path,
-                    "user": user,
-                    "query": self.text
-                })
+                log.add_debug(("Filtered out excluded search result %s from user %s for "
+                               'search term "%s"'), (file_path, user, self.text))
                 continue
 
             if not all(word in file_path_lower for word in search.included_words):
@@ -847,9 +842,8 @@ class Search:
                     size,
                     bitrate,
                     length,
-                    file_path,
                     has_free_slots,
-                    file_attributes,
+                    SearchResultFile(file_path, file_attributes),
                     row_id
                 ]
             )
@@ -866,8 +860,13 @@ class Search:
         if user in self.users:
             return
 
+        self.initialized = True
+
         ip_address, _port = msg.addr
-        country_code = core.network_filter.get_country_code(ip_address)
+        country_code = (
+            core.network_filter.get_country_code(ip_address)
+            or core.users.countries.get(user)
+        )
         has_free_slots = msg.freeulslots
 
         if has_free_slots:
@@ -928,20 +927,23 @@ class Search:
         return True
 
     def add_row_to_model(self, row):
-        (user, flag, h_speed, h_queue, folder_path, _unused, _unused, _unused, _unused,
-            _unused, speed, queue, _unused, _unused, _unused, file_path, has_free_slots,
-            _unused, _unused) = row
 
+        (user, flag, h_speed, h_queue, folder_path, _unused, _unused, _unused, _unused,
+            _unused, speed, queue, _unused, _unused, _unused, has_free_slots,
+            file_data, _unused) = row
+
+        expand_allowed = self.initialized
         expand_user = False
         expand_folder = False
         parent_iterator = None
+        user_child_iterators = None
+        user_folder_child_iterators = None
 
         if self.grouping_mode != "ungrouped":
             # Group by folder or user
 
             empty_int = 0
             empty_str = ""
-            empty_dict = {}
 
             if user not in self.users:
                 iterator = self.tree_view.add_row(
@@ -961,17 +963,14 @@ class Search:
                         empty_int,
                         empty_int,
                         empty_int,
-                        empty_str,
                         has_free_slots,
-                        empty_dict,
+                        None,
                         self.row_id
                     ], select_row=False
                 )
 
-                if self.grouping_mode == "folder_grouping":
-                    expand_user = True
-                else:
-                    expand_user = self.expand_button.get_active()
+                if expand_allowed:
+                    expand_user = self.grouping_mode == "folder_grouping" or self.expand_button.get_active()
 
                 self.row_id += 1
                 self.users[user] = (iterator, [])
@@ -1001,14 +1000,13 @@ class Search:
                             empty_int,
                             empty_int,
                             empty_int,
-                            file_path.rsplit("\\", 1)[0],
                             has_free_slots,
-                            empty_dict,
+                            SearchResultFile(file_data.path.rpartition("\\")[0]),
                             self.row_id
                         ], select_row=False, parent_iterator=user_iterator
                     )
                     user_child_iterators.append(iterator)
-                    expand_folder = self.expand_button.get_active()
+                    expand_folder = expand_allowed and self.expand_button.get_active()
                     self.row_id += 1
                     self.folders[user_folder_path] = (iterator, [])
 
@@ -1027,11 +1025,11 @@ class Search:
 
             user_iterator, user_child_iterators = self.users[user]
 
-        row[18] = self.row_id
+        row[17] = self.row_id
         iterator = self.tree_view.add_row(row, select_row=False, parent_iterator=parent_iterator)
         self.row_id += 1
 
-        if self.grouping_mode == "folder_grouping":
+        if user_folder_child_iterators is not None:
             user_folder_child_iterators.append(iterator)
         else:
             user_child_iterators.append(iterator)
@@ -1234,16 +1232,16 @@ class Search:
             if not filter_value:
                 continue
 
-            if filter_id == "filtertype" and not self.check_file_type(filter_value, row[15].lower()):
+            if filter_id == "filtertype" and not self.check_file_type(filter_value, row[16].path.lower()):
                 return False
 
             if filter_id == "filtercc" and not self.check_country(filter_value, row[1][-2:].upper()):
                 return False
 
-            if filter_id == "filterin" and not filter_value.search(row[15]) and not filter_value.fullmatch(row[0]):
+            if filter_id == "filterin" and not filter_value.search(row[16].path) and not filter_value.fullmatch(row[0]):
                 return False
 
-            if filter_id == "filterout" and (filter_value.search(row[15]) or filter_value.fullmatch(row[0])):
+            if filter_id == "filterout" and (filter_value.search(row[16].path) or filter_value.fullmatch(row[0])):
                 return False
 
             if filter_id == "filterslot" and row[11] > 0:
@@ -1271,6 +1269,8 @@ class Search:
 
     def clear_model(self, stored_results=False):
 
+        self.initialized = False
+
         if stored_results:
             self.all_data.clear()
             self.num_results_found = 0
@@ -1285,7 +1285,7 @@ class Search:
 
     def update_model(self):
 
-        self.tree_view.disable_sorting()
+        self.tree_view.freeze()
 
         for row in self.all_data:
             if self.check_filter(row):
@@ -1294,7 +1294,7 @@ class Search:
         # Update number of results
         self.update_result_counter()
 
-        self.tree_view.enable_sorting()
+        self.tree_view.unfreeze()
 
         if self.grouping_mode != "ungrouped":
             # Group by folder or user
@@ -1307,6 +1307,8 @@ class Search:
                 if self.grouping_mode == "folder_grouping":
                     self.tree_view.expand_root_rows()
 
+        self.initialized = True
+
     def update_wish_button(self):
 
         if self.mode not in {"global", "wishlist"}:
@@ -1347,9 +1349,9 @@ class Search:
         if not self.selected_users:
             return
 
-        # Multiple users, create submenus for each user
+        # Multiple users, create submenus for some of them
         if len(self.selected_users) > 1:
-            for user in self.selected_users:
+            for user in islice(self.selected_users, 20):
                 popup = UserPopupMenu(self.window.application, username=user, tab_name="search")
                 self.add_popup_menu_user(popup, user)
                 self.popup_menu_users.add_items((">" + user, popup))
@@ -1377,15 +1379,15 @@ class Search:
     def on_file_properties_accelerator(self, *_args):
         """Alt+Return - show file properties dialog."""
 
+        self.select_results()
         self.on_file_properties()
         return True
 
-    def on_select_user_results(self, *args):
+    def on_select_user_results(self, _action, _parameter, selected_user):
 
         if not self.selected_users:
             return
 
-        selected_user = args[-1]
         _user_iterator, user_child_iterators = self.users[selected_user]
 
         self.tree_view.unselect_all_rows()
@@ -1469,7 +1471,13 @@ class Search:
         self.results_label.set_text(humanize(self.num_results_visible) + str_plus)
 
     def on_file_path_tooltip(self, treeview, iterator):
-        return treeview.get_row_value(iterator, "file_path_data")
+
+        file_data = treeview.get_row_value(iterator, "file_data")
+
+        if not file_data:
+            return None
+
+        return file_data.path
 
     def on_row_activated(self, treeview, iterator, _column):
 
@@ -1503,9 +1511,9 @@ class Search:
             return
 
         user = self.tree_view.get_row_value(iterator, "user")
-        folder_path = self.tree_view.get_row_value(iterator, "file_path_data").rsplit("\\", 1)[0] + "\\"
+        path = self.tree_view.get_row_value(iterator, "file_data").path
 
-        core.userbrowse.browse_user(user, path=folder_path)
+        core.userbrowse.browse_user(user, path=path)
 
     def on_user_profile(self, *_args):
 
@@ -1524,20 +1532,13 @@ class Search:
         selected_length = 0
 
         for iterator in self.selected_results.values():
-            file_path = self.tree_view.get_row_value(iterator, "file_path_data")
+            file_data = self.tree_view.get_row_value(iterator, "file_data")
+            file_path = file_data.path
             file_size = self.tree_view.get_row_value(iterator, "size_data")
             selected_size += file_size
             selected_length += self.tree_view.get_row_value(iterator, "length_data")
             country_code = self.tree_view.get_row_value(iterator, "country")[-2:].upper()
-            country_name = core.network_filter.COUNTRIES.get(country_code, _("Unknown"))
-            country = f"{country_name} ({country_code})"
-
-            try:
-                folder_path, basename = file_path.rsplit("\\", 1)
-
-            except ValueError:
-                folder_path = ""
-                basename = file_path
+            folder_path, _separator, basename = file_path.rpartition("\\")
 
             data.append({
                 "user": self.tree_view.get_row_value(iterator, "user"),
@@ -1547,8 +1548,8 @@ class Search:
                 "size": file_size,
                 "speed": self.tree_view.get_row_value(iterator, "speed_data"),
                 "queue_position": self.tree_view.get_row_value(iterator, "in_queue_data"),
-                "file_attributes": self.tree_view.get_row_value(iterator, "file_attributes_data"),
-                "country": country
+                "file_attributes": file_data.attributes,
+                "country_code": country_code
             })
 
         if data:
@@ -1562,15 +1563,16 @@ class Search:
 
         for iterator in self.selected_results.values():
             user = self.tree_view.get_row_value(iterator, "user")
-            file_path = self.tree_view.get_row_value(iterator, "file_path_data")
+            file_data = self.tree_view.get_row_value(iterator, "file_data")
+            file_path = file_data.path
             size = self.tree_view.get_row_value(iterator, "size_data")
-            file_attributes = self.tree_view.get_row_value(iterator, "file_attributes_data")
 
             core.downloads.enqueue_download(
-                user, file_path, folder_path=download_folder_path, size=size, file_attributes=file_attributes)
+                user, file_path, folder_path=download_folder_path, size=size,
+                file_attributes=file_data.attributes)
 
-    def on_download_files_to_selected(self, selected_folder_path, _data):
-        self.on_download_files(download_folder_path=selected_folder_path)
+    def on_download_files_to_selected(self, selected_folder_paths, _data):
+        self.on_download_files(download_folder_path=next(iter(selected_folder_paths), None))
 
     def on_download_files_to(self, *_args):
 
@@ -1587,7 +1589,7 @@ class Search:
 
         for iterator in self.selected_results.values():
             user = self.tree_view.get_row_value(iterator, "user")
-            folder_path = self.tree_view.get_row_value(iterator, "file_path_data").rsplit("\\", 1)[0]
+            folder_path = self.tree_view.get_row_value(iterator, "file_data").path.rpartition("\\")[0]
             user_folder_key = user + folder_path
 
             if user_folder_key in requested_folders:
@@ -1598,22 +1600,22 @@ class Search:
             visible_files = []
             for row in self.all_data:
                 # Find the wanted folder
-                if folder_path != row[15].rsplit("\\", 1)[0]:
+                if folder_path != row[16].path.rpartition("\\")[0]:
                     continue
 
                 (_unused, _unused, _unused, _unused, _unused, _unused, _unused, _unused, _unused,
-                    _unused, _unused, _unused, size, _unused, _unused, file_path, _unused,
-                    file_attributes, _unused) = row
+                    _unused, _unused, _unused, size, _unused, _unused, _unused, file_data,
+                    _unused) = row
 
-                visible_files.append((file_path, size, file_attributes))
+                visible_files.append((file_data.path, size, file_data.attributes))
 
             core.search.request_folder_download(
                 user, folder_path, visible_files, download_folder_path=download_folder_path
             )
             requested_folders.add(user_folder_key)
 
-    def on_download_folders_to_selected(self, selected_folder_path, _data):
-        self.on_download_folders(download_folder_path=selected_folder_path)
+    def on_download_folders_to_selected(self, selected_folder_paths, _data):
+        self.on_download_folders(download_folder_path=next(iter(selected_folder_paths), None))
 
     def on_download_folders_to(self, *_args):
 
@@ -1631,7 +1633,7 @@ class Search:
         if iterator is None:
             return
 
-        file_path = self.tree_view.get_row_value(iterator, "file_path_data")
+        file_path = self.tree_view.get_row_value(iterator, "file_data").path
         clipboard.copy_text(file_path)
 
     def on_copy_url(self, *_args):
@@ -1642,7 +1644,7 @@ class Search:
             return
 
         user = self.tree_view.get_row_value(iterator, "user")
-        file_path = self.tree_view.get_row_value(iterator, "file_path_data")
+        file_path = self.tree_view.get_row_value(iterator, "file_data").path
         url = core.userbrowse.get_soulseek_url(user, file_path)
         clipboard.copy_text(url)
 
@@ -1654,8 +1656,10 @@ class Search:
             return
 
         user = self.tree_view.get_row_value(iterator, "user")
-        file_path = self.tree_view.get_row_value(iterator, "file_path_data")
-        url = core.userbrowse.get_soulseek_url(user, file_path.rsplit("\\", 1)[0] + "\\")
+        file_path = self.tree_view.get_row_value(iterator, "file_data").path
+        folder_path, separator, _basename = file_path.rpartition("\\")
+        url = core.userbrowse.get_soulseek_url(user, folder_path + separator)
+
         clipboard.copy_text(url)
 
     def on_counter_button(self, *_args):
