@@ -29,7 +29,6 @@ from gi.repository import Gtk
 
 from pynicotine.config import config
 from pynicotine.core import core
-from pynicotine.events import events
 from pynicotine.gtkgui.application import GTK_API_VERSION
 from pynicotine.gtkgui.dialogs.fileproperties import FileProperties
 from pynicotine.gtkgui.widgets import clipboard
@@ -102,7 +101,7 @@ class Transfers:
     UNKNOWN_STATUS_PRIORITY = 1000
 
     path_separator = path_label = retry_label = abort_label = None
-    transfer_page = user_counter = file_counter = expand_button = expand_icon = grouping_button = None
+    transfer_page = user_counter = file_counter = expand_button = expand_icon = grouping_button = status_label = None
 
     def __init__(self, window, transfer_type):
 
@@ -129,7 +128,7 @@ class Transfers:
         self.grouping_mode = None
         self.row_id = 0
         self.file_properties = None
-        self.pending_parent_rows_timer_id = None
+        self.initialized = False
 
         # Use dict instead of list for faster membership checks
         self.selected_users = {}
@@ -144,32 +143,37 @@ class Transfers:
                 "user": {
                     "column_type": "text",
                     "title": _("User"),
-                    "width": 200
+                    "width": 200,
+                    "sensitive_column": "is_sensitive_data"
                 },
                 "path": {
                     "column_type": "text",
                     "title": self.path_label,
                     "width": 200,
                     "expand_column": True,
-                    "tooltip_callback": self.on_file_path_tooltip
+                    "tooltip_callback": self.on_file_path_tooltip,
+                    "sensitive_column": "is_sensitive_data"
                 },
                 "file_type": {
                     "column_type": "icon",
                     "title": _("File Type"),
                     "width": 40,
-                    "hide_header": True
+                    "hide_header": True,
+                    "sensitive_column": "is_sensitive_data"
                 },
                 "filename": {
                     "column_type": "text",
                     "title": _("Filename"),
                     "width": 200,
                     "expand_column": True,
-                    "tooltip_callback": self.on_file_path_tooltip
+                    "tooltip_callback": self.on_file_path_tooltip,
+                    "sensitive_column": "is_sensitive_data"
                 },
                 "status": {
                     "column_type": "text",
                     "title": _("Status"),
-                    "width": 140
+                    "width": 140,
+                    "sensitive_column": "is_sensitive_data"
                 },
                 "queue_position": {
                     "column_type": "number",
@@ -180,31 +184,36 @@ class Transfers:
                 "percent": {
                     "column_type": "progress",
                     "title": _("Percent"),
-                    "width": 90
+                    "width": 90,
+                    "sensitive_column": "is_sensitive_data"
                 },
                 "size": {
                     "column_type": "number",
                     "title": _("Size"),
                     "width": 180,
-                    "sort_column": "size_data"
+                    "sort_column": "size_data",
+                    "sensitive_column": "is_sensitive_data"
                 },
                 "speed": {
                     "column_type": "number",
                     "title": _("Speed"),
                     "width": 100,
-                    "sort_column": "speed_data"
+                    "sort_column": "speed_data",
+                    "sensitive_column": "is_sensitive_data"
                 },
                 "time_elapsed": {
                     "column_type": "number",
                     "title": _("Time Elapsed"),
                     "width": 140,
-                    "sort_column": "time_elapsed_data"
+                    "sort_column": "time_elapsed_data",
+                    "sensitive_column": "is_sensitive_data"
                 },
                 "time_left": {
                     "column_type": "number",
                     "title": _("Time Left"),
                     "width": 140,
-                    "sort_column": "time_left_data"
+                    "sort_column": "time_left_data",
+                    "sensitive_column": "is_sensitive_data"
                 },
 
                 # Hidden data columns
@@ -212,11 +221,12 @@ class Transfers:
                 "current_bytes_data": {"data_type": GObject.TYPE_UINT64},
                 "speed_data": {"data_type": GObject.TYPE_UINT64},
                 "queue_position_data": {"data_type": GObject.TYPE_UINT},
-                "time_elapsed_data": {"data_type": int},
+                "time_elapsed_data": {"data_type": GObject.TYPE_INT},
                 "time_left_data": {"data_type": GObject.TYPE_UINT64},
+                "is_sensitive_data": {"data_type": GObject.TYPE_BOOLEAN},
                 "transfer_data": {"data_type": GObject.TYPE_PYOBJECT},
                 "id_data": {
-                    "data_type": GObject.TYPE_UINT64,
+                    "data_type": GObject.TYPE_INT,
                     "default_sort_type": "ascending",
                     "iterator_key": True
                 }
@@ -366,6 +376,16 @@ class Transfers:
 
         return status
 
+    def update_limits(self):
+        """Underline status bar bandwidth labels when alternative speed limits
+        are active."""
+
+        if config.sections["transfers"][f"use_{self.type}_speed_limit"] == "alternative":
+            add_css_class(self.status_label, "underline")
+            return
+
+        remove_css_class(self.status_label, "underline")
+
     def update_num_users_files(self):
         self.user_counter.set_text(humanize(len(self.users)))
         self.file_counter.set_text(humanize(len(self.transfer_list)))
@@ -380,6 +400,7 @@ class Transfers:
             # No need to do unnecessary work if transfers are not visible
             return
 
+        has_disabled_sorting = False
         has_selected_parent = False
         update_counters = False
         use_reverse_file_path = config.sections["ui"]["reverse_file_paths"]
@@ -395,8 +416,15 @@ class Transfers:
                 if select_parent:
                     has_selected_parent = True
 
-                if row_added:
-                    update_counters = True
+                if not row_added:
+                    continue
+
+                update_counters = True
+
+                if not has_disabled_sorting:
+                    # Optimization: disable sorting while adding rows
+                    self.tree_view.freeze()
+                    has_disabled_sorting = True
 
         if update_parent:
             self.update_parent_rows(transfer)
@@ -404,12 +432,12 @@ class Transfers:
         if update_counters:
             self.update_num_users_files()
 
-        if self.pending_parent_rows_timer_id is None:
-            # Limit individual parent row updates to once per second
-            self.pending_parent_rows_timer_id = events.schedule(
-                delay=1, callback=self._update_pending_parent_rows, repeat=True)
+        if has_disabled_sorting:
+            self.tree_view.unfreeze()
 
+        if not self.initialized:
             self.on_expand_tree()
+            self.initialized = True
 
         self.tree_view.redraw()
 
@@ -461,6 +489,10 @@ class Transfers:
 
     @staticmethod
     def get_hsize(current_byte_offset, size):
+
+        if current_byte_offset >= size:
+            return human_size(size)
+
         return f"{human_size(current_byte_offset)} / {human_size(size)}"
 
     @staticmethod
@@ -545,6 +577,13 @@ class Transfers:
 
         if transfer.status != parent_status:
             self.tree_view.set_row_value(iterator, "status", self.translate_status(parent_status))
+
+            if parent_status == TransferStatus.USER_LOGGED_OFF:
+                self.tree_view.set_row_value(iterator, "is_sensitive_data", False)
+
+            elif transfer.status == TransferStatus.USER_LOGGED_OFF:
+                self.tree_view.set_row_value(iterator, "is_sensitive_data", True)
+
             transfer.status = parent_status
 
         if transfer.speed != speed:
@@ -594,10 +633,17 @@ class Transfers:
         # Modify old transfer
         if iterator and iterator not in self.PENDING_ITERATORS:
             should_update_size = False
+            old_translated_status = self.tree_view.get_row_value(iterator, "status")
 
-            if self.tree_view.get_row_value(iterator, "status") != translated_status:
+            if old_translated_status != translated_status:
                 self.tree_view.set_row_value(iterator, "status", translated_status)
 
+                if transfer.status == TransferStatus.USER_LOGGED_OFF:
+                    self.tree_view.set_row_value(iterator, "is_sensitive_data", False)
+
+                elif old_translated_status == _("User logged off"):
+                    self.tree_view.set_row_value(iterator, "is_sensitive_data", True)
+
             if self.tree_view.get_row_value(iterator, "speed_data") != speed:
                 self.tree_view.set_row_value(iterator, "speed", self.get_hspeed(speed))
                 self.tree_view.set_row_value(iterator, "speed_data", speed)
@@ -626,16 +672,18 @@ class Transfers:
 
             return False
 
-        expand_allowed = self.pending_parent_rows_timer_id is not None
+        expand_allowed = self.initialized
         expand_user = False
         expand_folder = False
         user_iterator = None
         user_folder_path_iterator = None
         parent_iterator = None
+        select_iterator = None
 
         user = transfer.username
         folder_path, _separator, basename = transfer.virtual_path.rpartition("\\")
         original_folder_path = folder_path = self.get_transfer_folder_path(transfer)
+        is_sensitive = (status != TransferStatus.USER_LOGGED_OFF)
 
         if use_reverse_file_path:
             parts = folder_path.split(self.path_separator)
@@ -649,8 +697,6 @@ class Transfers:
         if self.grouping_mode != "ungrouped":
             # Group by folder or user
 
-            expand_user = expand_allowed
-            select_iterator = None
             empty_int = 0
             empty_str = ""
 
@@ -675,18 +721,23 @@ class Transfers:
                         empty_int,
                         empty_int,
                         empty_int,
+                        is_sensitive,
                         Transfer(user, status=status),  # Dummy Transfer object
                         self.row_id
                     ], select_row=False
                 )
 
-                if expand_allowed and not self.expand_button.get_active():
-                    expand_user = (self.grouping_mode == "folder_grouping")
+                if expand_allowed:
+                    expand_user = self.grouping_mode == "folder_grouping" or self.expand_button.get_active()
 
                 self.row_id += 1
                 self.users[user] = (iterator, [])
 
             user_iterator, user_child_transfers = self.users[user]
+            parent_iterator = user_iterator
+
+            if select_parent:
+                select_iterator = parent_iterator
 
             if self.grouping_mode == "folder_grouping":
                 # Group by folder
@@ -717,6 +768,7 @@ class Transfers:
                             empty_int,
                             empty_int,
                             empty_int,
+                            is_sensitive,
                             path_transfer,
                             self.row_id
                         ], select_row=False, parent_iterator=user_iterator
@@ -730,28 +782,13 @@ class Transfers:
                 parent_iterator = user_folder_path_iterator
                 user_folder_path_child_transfers.append(transfer)
 
-                if select_parent:
-                    self.tree_view.expand_row(user_iterator)
-                    select_iterator = user_folder_path_iterator
-                    expand_user = False
+                if select_parent and (expand_user or self.tree_view.is_row_expanded(user_iterator)):
+                    select_iterator = parent_iterator
 
                 # Group by folder, path not visible in file rows
                 folder_path = ""
             else:
-                parent_iterator = user_iterator
                 user_child_transfers.append(transfer)
-
-                if select_parent:
-                    select_iterator = user_iterator
-
-            if select_iterator and (not self.tree_view.is_row_selected(select_iterator)
-                                    or self.tree_view.get_num_selected_rows() != 1):
-                # Select parent row of newly added transfer, and scroll to it.
-                # Unselect any other rows to prevent accidental actions on previously
-                # selected transfers.
-                self.tree_view.unselect_all_rows()
-                self.tree_view.select_row(select_iterator, expand_rows=False)
-
         else:
             # No grouping
             if user not in self.users:
@@ -779,6 +816,7 @@ class Transfers:
             queue_position,
             elapsed,
             left,
+            is_sensitive,
             transfer,
             self.row_id
         ], select_row=False, parent_iterator=parent_iterator)
@@ -790,14 +828,19 @@ class Transfers:
         if expand_folder and user_folder_path_iterator is not None:
             self.tree_view.expand_row(user_folder_path_iterator)
 
+        if select_iterator and (not self.tree_view.is_row_selected(select_iterator)
+                                or self.tree_view.get_num_selected_rows() != 1):
+            # Select parent row of newly added transfer, and scroll to it.
+            # Unselect any other rows to prevent accidental actions on previously
+            # selected transfers.
+            self.tree_view.unselect_all_rows()
+            self.tree_view.select_row(select_iterator, expand_rows=False)
+
         return True
 
     def clear_model(self):
 
-        if self.pending_parent_rows_timer_id is not None:
-            events.cancel_scheduled(self.pending_parent_rows_timer_id)
-            self.pending_parent_rows_timer_id = None
-
+        self.initialized = False
         self.users.clear()
         self.paths.clear()
         self.pending_folder_rows.clear()
@@ -992,12 +1035,11 @@ class Transfers:
         elif action == 7:  # Browse Folder
             self.on_browse_folder()
 
-    def on_select_user_transfers(self, *args):
+    def on_select_user_transfers(self, _action, _parameter, selected_user):
 
         if not self.selected_users:
             return
 
-        selected_user = args[-1]
         _user_iterator, user_child_transfers = self.users[selected_user]
 
         self.tree_view.unselect_all_rows()
@@ -1063,6 +1105,9 @@ class Transfers:
         selected_length = 0
 
         for transfer in self.selected_transfers:
+            username = transfer.username
+            watched_user = core.users.watched.get(username)
+            speed = 0
             file_path = transfer.virtual_path
             file_size = transfer.size
             file_attributes = transfer.file_attributes
@@ -1074,6 +1119,9 @@ class Transfers:
 
             folder_path, _separator, basename = file_path.rpartition("\\")
 
+            if watched_user is not None:
+                speed = watched_user.upload_speed or 0
+
             data.append({
                 "user": transfer.username,
                 "file_path": file_path,
@@ -1081,9 +1129,10 @@ class Transfers:
                 "virtual_folder_path": folder_path,
                 "real_folder_path": transfer.folder_path,
                 "queue_position": transfer.queue_position,
-                "speed": transfer.speed,
+                "speed": speed,
                 "size": file_size,
-                "file_attributes": file_attributes
+                "file_attributes": file_attributes,
+                "country_code": core.users.countries.get(username)
             })
 
         if data:
