@@ -23,8 +23,6 @@
 
 import os
 
-from locale import strxfrm
-
 from gi.repository import GLib
 from gi.repository import GObject
 from gi.repository import Gtk
@@ -38,6 +36,7 @@ from pynicotine.gtkgui.dialogs.fileproperties import FileProperties
 from pynicotine.gtkgui.widgets import clipboard
 from pynicotine.gtkgui.widgets import ui
 from pynicotine.gtkgui.widgets.accelerator import Accelerator
+from pynicotine.gtkgui.widgets.combobox import ComboBox
 from pynicotine.gtkgui.widgets.dialogs import EntryDialog
 from pynicotine.gtkgui.widgets.filechooser import FolderChooser
 from pynicotine.gtkgui.widgets.iconnotebook import IconNotebook
@@ -45,13 +44,13 @@ from pynicotine.gtkgui.widgets.infobar import InfoBar
 from pynicotine.gtkgui.widgets.popupmenu import PopupMenu
 from pynicotine.gtkgui.widgets.popupmenu import FilePopupMenu
 from pynicotine.gtkgui.widgets.popupmenu import UserPopupMenu
-from pynicotine.gtkgui.widgets.textentry import ComboBox
 from pynicotine.gtkgui.widgets.theme import add_css_class
 from pynicotine.gtkgui.widgets.theme import get_file_type_icon_name
 from pynicotine.gtkgui.widgets.theme import remove_css_class
 from pynicotine.gtkgui.widgets.treeview import TreeView
-from pynicotine.slskmessages import UserStatus
+from pynicotine.slskmessages import ConnectionType
 from pynicotine.slskmessages import FileListMessage
+from pynicotine.slskmessages import UserStatus
 from pynicotine.utils import human_size
 from pynicotine.utils import humanize
 from pynicotine.utils import open_file_path
@@ -88,7 +87,6 @@ class UserBrowses(IconNotebook):
             ("peer-connection-error", self.peer_connection_error),
             ("quit", self.quit),
             ("server-disconnect", self.server_disconnect),
-            ("server-login", self.on_focus),
             ("shared-file-list-progress", self.shared_file_list_progress),
             ("shared-file-list-response", self.shared_file_list),
             ("user-browse-remove-user", self.remove_user),
@@ -174,11 +172,14 @@ class UserBrowses(IconNotebook):
         del self.pages[user]
         page.destroy()
 
-    def peer_connection_error(self, user, *_args, **_kwargs):
+    def peer_connection_error(self, username, conn_type, **_unused):
 
-        page = self.pages.get(user)
+        page = self.pages.get(username)
 
-        if page is not None:
+        if page is None:
+            return
+
+        if conn_type == ConnectionType.PEER:
             page.peer_connection_error()
 
     def user_status(self, msg):
@@ -203,9 +204,7 @@ class UserBrowses(IconNotebook):
             page.shared_file_list(msg)
 
     def server_disconnect(self, *_args):
-
         for user, page in self.pages.items():
-            page.peer_connection_error()
             self.set_user_status(page.container, user, UserStatus.OFFLINE)
 
 
@@ -390,10 +389,6 @@ class UserBrowse:
             )
 
         # Key Bindings (folder_tree_view)
-        Accelerator("Left", self.folder_tree_view.widget, self.on_folder_collapse_accelerator)
-        Accelerator("minus", self.folder_tree_view.widget, self.on_folder_collapse_accelerator)  # "-"
-        Accelerator("backslash", self.folder_tree_view.widget, self.on_folder_collapse_sub_accelerator)  # "\"
-        Accelerator("equal", self.folder_tree_view.widget, self.on_folder_expand_sub_accelerator)  # "=" (for US/UK)
         Accelerator("Right", self.folder_tree_view.widget, self.on_folder_expand_accelerator)
 
         Accelerator("<Shift>Return", self.folder_tree_view.widget, self.on_folder_focus_filetree_accelerator)
@@ -403,7 +398,7 @@ class UserBrowse:
         Accelerator("<Alt>Return", self.folder_tree_view.widget, self.on_file_properties_accelerator, True)
 
         # Key Bindings (file_list_view)
-        for accelerator in ("<Shift>Tab", "BackSpace", "backslash"):  # Avoid header, navigate up, "\"
+        for accelerator in ("BackSpace", "backslash"):  # Navigate up, "\"
             Accelerator(accelerator, self.file_list_view.widget, self.on_focus_folder_accelerator)
 
         Accelerator("Left", self.file_list_view.widget, self.on_focus_folder_left_accelerator)
@@ -423,6 +418,8 @@ class UserBrowse:
         Accelerator("<Shift>F3", self.container, self.on_search_previous_accelerator)
         Accelerator("<Primary>g", self.container, self.on_search_next_accelerator)  # Next search match
         Accelerator("<Shift><Primary>g", self.container, self.on_search_previous_accelerator)
+        Accelerator("Up", self.search_entry, self.on_search_previous_accelerator)
+        Accelerator("Down", self.search_entry, self.on_search_next_accelerator)
 
         Accelerator("<Primary>backslash", self.container, self.on_expand_accelerator)  # expand / collapse all (button)
         Accelerator("F5", self.container, self.on_refresh_accelerator)
@@ -470,18 +467,19 @@ class UserBrowse:
     def rebuild_model(self):
 
         self.clear_model()
-
-        private_size = num_private_folders = 0
         browsed_user = core.userbrowse.users[self.user]
 
+        if browsed_user.num_folders is None or browsed_user.shared_size is None:
+            return
+
         # Generate the folder tree and select first folder
-        size, num_folders = self.create_folder_tree(browsed_user.public_folders)
+        self.create_folder_tree(browsed_user.public_folders)
 
         if browsed_user.private_folders:
-            private_size, num_private_folders = self.create_folder_tree(browsed_user.private_folders, private=True)
+            self.create_folder_tree(browsed_user.private_folders, private=True)
 
-        self.num_folders_label.set_text(humanize(num_folders + num_private_folders))
-        self.share_size_label.set_text(human_size(size + private_size))
+        self.num_folders_label.set_text(humanize(browsed_user.num_folders))
+        self.share_size_label.set_text(human_size(browsed_user.shared_size))
 
         if self.expand_button.get_active():
             self.folder_tree_view.expand_all_rows()
@@ -492,37 +490,37 @@ class UserBrowse:
 
     def create_folder_tree(self, folders, private=False):
 
-        total_size = 0
-        num_folders = len(folders)
-
         if not folders:
-            return total_size, num_folders
+            return
 
-        for folder_path, files in folders.items():
-            current_path = None
+        iterators = self.folder_tree_view.iterators
+        add_row = self.folder_tree_view.add_row
+        query = self.query
+        private_template = _("[PRIVATE]  %s")
+
+        for folder_path, files in reversed(list(folders.items())):
+            current_path = parent = None
             root_processed = False
-            skip_folder = (self.query and self.query not in folder_path.lower())
+            skip_folder = (query and query not in folder_path.lower())
 
-            for filedata in files:
-                if skip_folder and self.query in filedata[1].lower():
-                    skip_folder = False
-
-                total_size += filedata[2]
+            if skip_folder:
+                for file_info in files:
+                    if query in file_info[1].lower():
+                        skip_folder = False
 
             if skip_folder:
                 continue
 
             for subfolder in folder_path.split("\\"):
-                parent = self.folder_tree_view.iterators.get(current_path)
-
                 if not root_processed:
                     current_path = subfolder
                     root_processed = True
                 else:
-                    current_path = "\\".join([current_path, subfolder])
+                    current_path += f"\\{subfolder}"
 
-                if current_path in self.folder_tree_view.iterators:
+                if current_path in iterators:
                     # Folder was already added to tree
+                    parent = iterators[current_path]
                     continue
 
                 if not subfolder:
@@ -530,16 +528,16 @@ class UserBrowse:
                     subfolder = "\\"
 
                 if private:
-                    subfolder = _("[PRIVATE]  %s") % subfolder
+                    subfolder = private_template % subfolder
 
-                self.folder_tree_view.add_row(
+                parent = add_row(
                     [subfolder, current_path], select_row=False, parent_iterator=parent
                 )
 
-            if self.query:
+            if query:
                 self.search_folder_paths.append(folder_path)
 
-        return total_size, num_folders
+        self.search_folder_paths.reverse()
 
     def browse_queued_path(self):
 
@@ -610,7 +608,23 @@ class UserBrowse:
         self.progress_bar.pulse()
         return repeat
 
-    def set_in_progress(self):
+    def shared_file_list_progress(self, position, total):
+
+        self.indeterminate_progress = False
+
+        if total <= 0 or position <= 0:
+            fraction = 0.0
+
+        elif position < total:
+            fraction = float(position) / total
+
+        else:
+            fraction = 1.0
+            GLib.timeout_add(1000, self.set_finishing)
+
+        self.progress_bar.set_fraction(fraction)
+
+    def set_indeterminate_progress(self):
 
         self.indeterminate_progress = True
 
@@ -624,18 +638,12 @@ class UserBrowse:
         self.refresh_button.set_sensitive(False)
         self.save_button.set_sensitive(False)
 
-    def shared_file_list_progress(self, position, total):
+    def set_finishing(self):
 
-        self.indeterminate_progress = False
+        if not self.refresh_button.get_sensitive():
+            self.set_indeterminate_progress()
 
-        if total <= 0 or position <= 0:
-            fraction = 0.0
-        elif position >= total:
-            fraction = 1.0
-        else:
-            fraction = float(position) / total
-
-        self.progress_bar.set_fraction(fraction)
+        return False
 
     def set_finished(self):
 
@@ -747,7 +755,7 @@ class UserBrowse:
                 return
 
         # Temporarily disable sorting for increased performance
-        self.file_list_view.disable_sorting()
+        self.file_list_view.freeze()
 
         for _code, basename, size, _ext, file_attributes, *_unused in files:
             h_size = human_size(size, config.sections["ui"]["file_size_unit"])
@@ -765,7 +773,7 @@ class UserBrowse:
                 file_attributes
             ], select_row=False)
 
-        self.file_list_view.enable_sorting()
+        self.file_list_view.unfreeze()
         self.select_search_match_files()
 
     def select_files(self):
@@ -828,11 +836,17 @@ class UserBrowse:
 
         if self.query != query:
             # New search query, rebuild result list
-            self.clear_model()
+            active_folder_path = self.active_folder_path
+
             self.query = query
             self.rebuild_model()
 
             if not self.search_folder_paths:
+                iterator = self.folder_tree_view.iterators.get(active_folder_path)
+
+                if iterator:
+                    self.folder_tree_view.select_row(iterator)
+
                 return False
 
         elif query:
@@ -862,9 +876,13 @@ class UserBrowse:
         if iterator is None:
             return
 
-        iterators = tree_view.get_selected_rows()
+        selected_iterators = tree_view.get_selected_rows()
+        folder_path = None
 
-        if len(iterators) > 1:
+        # Skip first folder
+        next(selected_iterators)
+
+        if next(selected_iterators, None):
             # Multiple folders selected. Avoid any confusion by clearing the path bar and file list view.
             folder_path = None
         else:
@@ -898,8 +916,9 @@ class UserBrowse:
     def on_download_folder_recursive(self, *_args):
         self.on_download_folder(recurse=True)
 
-    def on_download_folder_to_selected(self, selected_download_folder_path, recurse):
-        self.on_download_folder(download_folder_path=selected_download_folder_path, recurse=recurse)
+    def on_download_folder_to_selected(self, selected_download_folder_paths, recurse):
+        self.on_download_folder(
+            download_folder_path=next(iter(selected_download_folder_paths), None), recurse=recurse)
 
     def on_download_folder_to(self, *_args, recurse=False):
 
@@ -959,7 +978,7 @@ class UserBrowse:
             action_button_label=_("_Upload"),
             callback=self.on_upload_folder_to_response,
             callback_data=recurse,
-            droplist=sorted(core.buddies.users, key=strxfrm)
+            droplist=sorted(core.buddies.users)
         ).present()
 
     def on_upload_folder_recursive_to(self, *_args):
@@ -994,17 +1013,6 @@ class UserBrowse:
         # Note: Other Folder actions are handled by Accelerator functions [Shift/Ctrl/Alt+Return]
         # TODO: Mouse double-click actions will need keycode state & mods [Shift/Ctrl+DblClick]
 
-    def on_folder_collapse_accelerator(self, *_args):
-        """Left, Shift+Left (Gtk), "-", "/" (Gtk) - collapse row."""
-
-        iterator = self.folder_tree_view.get_focused_row()
-
-        if iterator is None:
-            return False
-
-        self.folder_tree_view.collapse_row(iterator)
-        return True
-
     def on_folder_expand_accelerator(self, *_args):
         """Right, Shift+Right (Gtk), "+" (Gtk) - expand row."""
 
@@ -1013,36 +1021,11 @@ class UserBrowse:
         if iterator is None:
             return False
 
-        expandable = self.folder_tree_view.expand_row(iterator)
-
-        if not expandable and not self.file_list_view.is_empty():
+        if not self.file_list_view.is_empty():
             self.file_list_view.grab_focus()
 
         return True
 
-    def on_folder_collapse_sub_accelerator(self, *_args):
-        """\backslash: collapse or expand to show subs."""
-
-        iterator = self.folder_tree_view.get_focused_row()
-
-        if iterator is None:
-            return False
-
-        self.folder_tree_view.collapse_row(iterator)  # show 2nd level
-        self.folder_tree_view.expand_row(iterator)
-        return True
-
-    def on_folder_expand_sub_accelerator(self, *_args):
-        """=equal: expand only (dont move focus)"""
-
-        iterator = self.folder_tree_view.get_focused_row()
-
-        if iterator is None:
-            return False
-
-        self.folder_tree_view.expand_row(iterator)
-        return True
-
     def on_folder_focus_filetree_accelerator(self, *_args):
         """Shift+Enter - focus selection over FileTree."""
 
@@ -1050,7 +1033,12 @@ class UserBrowse:
             self.file_list_view.grab_focus()
             return True
 
-        self.on_folder_expand_sub_accelerator()
+        iterator = self.folder_tree_view.get_focused_row()
+
+        if iterator is None:
+            return False
+
+        self.folder_tree_view.expand_row(iterator)
         return True
 
     def on_folder_transfer_to_accelerator(self, *_args):
@@ -1114,8 +1102,8 @@ class UserBrowse:
             core.userbrowse.download_file(
                 self.user, folder_path, file_data, download_folder_path=download_folder_path)
 
-    def on_download_files_to_selected(self, selected_download_folder_path, _data):
-        self.on_download_files(download_folder_path=selected_download_folder_path)
+    def on_download_files_to_selected(self, selected_download_folder_paths, _data):
+        self.on_download_files(download_folder_path=next(iter(selected_download_folder_paths), None))
 
     def on_download_files_to(self, *_args):
 
@@ -1147,7 +1135,7 @@ class UserBrowse:
             message=_("Enter the name of the user you want to upload to:"),
             action_button_label=_("_Upload"),
             callback=self.on_upload_files_to_response,
-            droplist=sorted(core.buddies.users, key=strxfrm)
+            droplist=sorted(core.buddies.users)
         ).present()
 
     def on_open_file(self, *_args):
@@ -1169,6 +1157,11 @@ class UserBrowse:
         data = []
         selected_size = 0
         selected_length = 0
+        watched_user = core.users.watched.get(self.user)
+        speed = 0
+
+        if watched_user is not None:
+            speed = watched_user.upload_speed or 0
 
         if all_files:
             prev_folder_path = None
@@ -1197,6 +1190,7 @@ class UserBrowse:
                             "file_path": file_path,
                             "basename": basename,
                             "virtual_folder_path": folder_path,
+                            "speed": speed,
                             "size": file_size,
                             "file_attributes": file_attributes
                         })
@@ -1218,8 +1212,10 @@ class UserBrowse:
                     "file_path": file_path,
                     "basename": basename,
                     "virtual_folder_path": selected_folder_path,
+                    "speed": speed,
                     "size": file_size,
-                    "file_attributes": self.file_list_view.get_row_value(iterator, "file_attributes_data")
+                    "file_attributes": self.file_list_view.get_row_value(iterator, "file_attributes_data"),
+                    "country_code": core.users.countries.get(self.user)
                 })
 
         if data:
@@ -1254,14 +1250,14 @@ class UserBrowse:
 
         column_id = self.file_list_view.get_focused_column()
 
-        if self.file_list_view.get_visible_columns()[0] != column_id:
+        if next(self.file_list_view.get_visible_columns(), None) != column_id:
             return False  # allow horizontal scrolling
 
         self.folder_tree_view.grab_focus()
         return True
 
     def on_focus_folder_accelerator(self, *_args):
-        """Shift+Tab, BackSpace, \backslash - focus selection back parent folder"""
+        """BackSpace, \backslash - focus selection back parent folder"""
 
         self.folder_tree_view.grab_focus()
         return True
@@ -1356,7 +1352,10 @@ class UserBrowse:
         """Enables indeterminate progress bar mode when tab is active."""
 
         if not self.indeterminate_progress and progress_bar.get_fraction() <= 0.0:
-            self.set_in_progress()
+            self.set_indeterminate_progress()
+
+        if core.users.login_status == UserStatus.OFFLINE:
+            self.peer_connection_error()
 
     def on_hide_progress_bar(self, progress_bar):
         """Disables indeterminate progress bar mode when switching to another tab."""
@@ -1405,11 +1404,18 @@ class UserBrowse:
     def on_show_search(self, *_args):
 
         active = self.search_button.get_active()
-        self.search_entry_revealer.set_reveal_child(active)
 
         if active:
             self.search_entry.grab_focus()
 
+        elif not self.file_list_view.is_selection_empty():
+            self.file_list_view.grab_focus()
+
+        else:
+            self.folder_tree_view.grab_focus()
+
+        self.search_entry_revealer.set_reveal_child(active)
+
     def on_search(self, *_args):
         self.find_search_matches()
 
@@ -1431,7 +1437,7 @@ class UserBrowse:
         file_path = self.get_selected_file_path()
 
         self.clear_model()
-        self.set_in_progress()
+        self.set_indeterminate_progress()
 
         if self.local_permission_level:
             core.userbrowse.browse_local_shares(
@@ -1506,10 +1512,4 @@ class UserBrowse:
         """Escape - navigate out of search_entry."""
 
         self.search_button.set_active(False)
-
-        if not self.file_list_view.is_selection_empty():
-            self.file_list_view.grab_focus()
-        else:
-            self.folder_tree_view.grab_focus()
-
         return True
