@@ -20,8 +20,6 @@
 # You should have received a copy of the GNU General Public License
 # along with this program.  If not, see <http://www.gnu.org/licenses/>.
 
-import random
-import string
 import time
 
 import gi.module
@@ -34,8 +32,6 @@ from gi.repository import Gtk
 from pynicotine.config import config
 from pynicotine.core import core
 from pynicotine.gtkgui.application import GTK_API_VERSION
-from pynicotine.gtkgui.application import GTK_MICRO_VERSION
-from pynicotine.gtkgui.application import GTK_MINOR_VERSION
 from pynicotine.gtkgui.widgets import clipboard
 from pynicotine.gtkgui.widgets.accelerator import Accelerator
 from pynicotine.gtkgui.widgets.popupmenu import PopupMenu
@@ -46,7 +42,7 @@ from pynicotine.gtkgui.widgets.theme import add_css_class
 
 class TreeView:
 
-    def __init__(self, window, parent, columns, has_tree=False, multi_select=False, always_select=False,
+    def __init__(self, window, parent, columns, has_tree=False, multi_select=False,
                  persistent_sort=False, name=None, secondary_name=None, activate_row_callback=None,
                  focus_in_callback=None, select_row_callback=None, delete_accelerator_callback=None,
                  search_entry=None):
@@ -65,7 +61,7 @@ class TreeView:
         self._iterator_key_column = 0
         self._column_ids = {}
         self._column_offsets = {}
-        self._column_gvalues = []
+        self._column_gvalues = {}
         self._column_gesture_controllers = []
         self._column_numbers = None
         self._default_sort_column = None
@@ -91,6 +87,15 @@ class TreeView:
         self._initialise_columns(columns)
 
         Accelerator("<Primary>c", self.widget, self.on_copy_cell_data_accelerator)
+        Accelerator("<Primary>f", self.widget, self.on_start_search)
+
+        Accelerator("Left", self.widget, self.on_collapse_row_accelerator)
+        Accelerator("minus", self.widget, self.on_collapse_row_blocked_accelerator)
+        Accelerator("Right", self.widget, self.on_expand_row_accelerator)
+        Accelerator("plus", self.widget, self.on_expand_row_blocked_accelerator)
+        Accelerator("equal", self.widget, self.on_expand_row_blocked_accelerator)
+        Accelerator("backslash", self.widget, self.on_expand_row_level_accelerator)
+
         self._column_menu = self.widget.column_menu = PopupMenu(
             self.window.application, self.widget, callback=self.on_column_header_menu, connect_events=False)
 
@@ -98,9 +103,6 @@ class TreeView:
             self.widget.set_rubber_banding(True)
             self._selection.set_mode(Gtk.SelectionMode.MULTIPLE)
 
-        elif always_select:
-            self._selection.set_mode(Gtk.SelectionMode.BROWSE)
-
         if activate_row_callback:
             self.widget.connect("row-activated", self.on_activate_row, activate_row_callback)
 
@@ -122,6 +124,7 @@ class TreeView:
             self.widget.set_search_entry(search_entry)
 
         self._query_tooltip_handler = self.widget.connect("query-tooltip", self.on_tooltip)
+        self.widget.connect("move-cursor", self.on_key_move_cursor)
         self.widget.set_search_equal_func(self.on_search_match)
 
         add_css_class(self.widget, "treeview-spacing")
@@ -220,6 +223,7 @@ class TreeView:
     def _initialise_column_ids(self, columns):
 
         self._data_types = []
+        int_types = {GObject.TYPE_UINT, GObject.TYPE_UINT64}
 
         for column_index, (column_id, column_data) in enumerate(columns.items()):
             data_type = column_data.get("data_type")
@@ -228,20 +232,19 @@ class TreeView:
                 column_type = column_data.get("column_type")
 
                 if column_type == "progress":
-                    data_type = int
+                    data_type = GObject.TYPE_INT
 
                 elif column_type == "toggle":
-                    data_type = bool
+                    data_type = GObject.TYPE_BOOLEAN
 
                 else:
-                    data_type = str
-
-            gvalue = GObject.Value(data_type)
-
-            self._data_types.append(data_type)
-            self._column_gvalues.append(gvalue)
+                    data_type = GObject.TYPE_STRING
 
             self._column_ids[column_id] = column_index
+            self._data_types.append(data_type)
+
+            if data_type in int_types:
+                self._column_gvalues[column_index] = GObject.Value(data_type)
 
         self._column_numbers = list(self._column_ids.values())
 
@@ -262,6 +265,7 @@ class TreeView:
             title = column_data.get("title")
             iterator_key = column_data.get("iterator_key")
             sort_data_column = column_data.get("sort_column", column_id)
+            sort_column_id = self._column_ids[sort_data_column]
             default_sort_type = column_data.get("default_sort_type")
 
             if iterator_key:
@@ -270,7 +274,7 @@ class TreeView:
 
             if default_sort_type:
                 # Sort treeview by values in this column by default
-                self._default_sort_column = self._column_ids[sort_data_column]
+                self._default_sort_column = sort_column_id
                 self._default_sort_type = (Gtk.SortType.DESCENDING if default_sort_type == "descending"
                                            else Gtk.SortType.ASCENDING)
 
@@ -305,7 +309,7 @@ class TreeView:
 
                 if column_sort_type and self._persistent_sort:
                     # Sort treeview by values in this column by default
-                    self._sort_column = self._column_ids[sort_data_column]
+                    self._sort_column = sort_column_id
                     self._sort_type = (Gtk.SortType.DESCENDING if column_sort_type == "descending"
                                        else Gtk.SortType.ASCENDING)
                     self.model.set_sort_column_id(self._sort_column, self._sort_type)
@@ -370,7 +374,7 @@ class TreeView:
                 gesture_click = Gtk.GestureMultiPress(widget=column_header)
 
             gesture_click.set_propagation_phase(Gtk.PropagationPhase.CAPTURE)
-            gesture_click.connect("released", self.on_column_header_pressed, column)
+            gesture_click.connect("released", self.on_column_header_pressed, column_id, sort_column_id)
             self._column_gesture_controllers.append(gesture_click)
 
             title_container = next(iter(column_header))
@@ -411,7 +415,7 @@ class TreeView:
             column.type = column_type
             column.tooltip_callback = column_data.get("tooltip_callback")
 
-            column.set_sort_column_id(self._column_ids[sort_data_column])
+            column.set_sort_column_id(sort_column_id)
             column_widgets[column_id] = column
 
         self.widget.set_headers_visible(has_visible_column_header)
@@ -464,51 +468,58 @@ class TreeView:
                 columns["sort"] = "descending" if self._sort_type == Gtk.SortType.DESCENDING else "ascending"
 
         if self._secondary_name is not None:
-            try:
-                column_config[self._widget_name]
-            except KeyError:
+            if self._widget_name not in column_config:
                 column_config[self._widget_name] = {}
 
             column_config[self._widget_name][self._secondary_name] = saved_columns
         else:
             column_config[self._widget_name] = saved_columns
 
-    def disable_sorting(self):
+    def freeze(self):
         self.model.set_sort_column_id(Gtk.TREE_SORTABLE_UNSORTED_SORT_COLUMN_ID, Gtk.SortType.ASCENDING)
 
-    def enable_sorting(self):
+    def unfreeze(self):
         if self._sort_column is not None and self._sort_type is not None:
             self.model.set_sort_column_id(self._sort_column, self._sort_type)
 
     def set_show_expanders(self, show):
         self.widget.set_show_expanders(show)
 
-    def add_row(self, values, select_row=True, prepend=False, parent_iterator=None):
+    def add_row(self, values, select_row=True, parent_iterator=None):
 
-        position = 0 if prepend else -1
-        values = values[:]
         key = values[self._iterator_key_column]
 
         if key in self.iterators:
             return None
 
-        for i, value in enumerate(values):
-            if isinstance(value, (float, int)) and value > 2147483647:
+        position = 0  # Insert at the beginning for large performance improvement
+        value_columns = []
+        included_values = []
+
+        for index, value in enumerate(values):
+            if not value and index is not self._sort_column:
+                # Skip empty values if not active sort column to avoid unnecessary work
+                continue
+
+            if index in self._column_gvalues:
                 # Need gvalue conversion for large integers
-                gvalue = self._column_gvalues[i]
-                gvalue.set_value(value)
+                gvalue = self._column_gvalues[index]
+                gvalue.set_value(value or 0)
                 value = gvalue
 
-            values[i] = value
+            value_columns.append(index)
+            included_values.append(value)
 
         if self.has_tree:
-            self.iterators[key] = iterator = self.model.insert_with_values(  # pylint: disable=no-member
-                parent_iterator, position, self._column_numbers, values
+            self.iterators[key] = iterator = self.model.insert_with_values(
+                parent_iterator, position, value_columns, included_values
             )
         else:
-            self.iterators[key] = iterator = self.model.insert_with_values(position, self._column_numbers, values)
+            self.iterators[key] = iterator = self.model.insert_with_values(
+                position, value_columns, included_values
+            )
 
-        self._iterator_keys[iterator.user_data] = key
+        self._iterator_keys[iterator] = key
 
         if select_row:
             self.select_row(iterator)
@@ -518,9 +529,9 @@ class TreeView:
     def get_selected_rows(self):
 
         _model, paths = self._selection.get_selected_rows()
-        iterators = [self.model.get_iter(path) for path in paths]
 
-        return iterators
+        for path in paths:
+            yield self.model.get_iter(path)
 
     def get_num_selected_rows(self):
         return self._selection.count_selected_rows()
@@ -541,7 +552,7 @@ class TreeView:
 
         column_index = self._column_ids[column_id]
 
-        if isinstance(value, (float, int)) and value > 2147483647:
+        if column_index in self._column_gvalues:
             # Need gvalue conversion for large integers
             gvalue = self._column_gvalues[column_index]
             gvalue.set_value(value)
@@ -550,7 +561,7 @@ class TreeView:
         return self.model.set_value(iterator, column_index, value)
 
     def remove_row(self, iterator):
-        del self.iterators[self._iterator_keys[iterator.user_data]]
+        del self.iterators[self._iterator_keys[iterator]]
         self.model.remove(iterator)
 
     def select_row(self, iterator=None, expand_rows=True, should_scroll=True):
@@ -575,6 +586,9 @@ class TreeView:
 
         self._selection.select_iter(iterator)
 
+    def select_all_rows(self):
+        self._selection.select_all()
+
     def unselect_all_rows(self):
         self._selection.unselect_all()
 
@@ -607,7 +621,10 @@ class TreeView:
         return column.id
 
     def get_visible_columns(self):
-        return [column.id for column in self.widget.get_columns() if column.get_visible()]
+
+        for column in self.widget.get_columns():
+            if column.get_visible():
+                yield column.id
 
     def is_empty(self):
         return not self.iterators
@@ -628,11 +645,13 @@ class TreeView:
     def clear(self):
 
         self.widget.set_model(None)
+        self.freeze()
 
         self.model.clear()
         self.iterators.clear()
         self._iterator_keys.clear()
 
+        self.unfreeze()
         self.widget.set_model(self.model)
 
     @staticmethod
@@ -671,13 +690,11 @@ class TreeView:
         callback(self)
 
     def on_select_row(self, selection, callback):
+
         iterator = None
 
         if self.multi_select:
-            # We use the bottom selection here.
-            iterators = self.get_selected_rows()
-            if len(iterators):
-                iterator = iterators[-1]
+            iterator = next(self.get_selected_rows(), None)
         else:
             _model, iterator = selection.get_selected()
 
@@ -686,7 +703,7 @@ class TreeView:
     def on_delete_accelerator(self, _treeview, _state, callback):
         callback(self)
 
-    def on_column_header_pressed(self, controller, _num_p, _pos_x, _pos_y, column):
+    def on_column_header_pressed(self, controller, _num_p, _pos_x, _pos_y, column_id, sort_column_id):
         """Reset sorting when column header has been pressed three times."""
 
         self._sort_column, self._sort_type = self.model.get_sort_column_id()
@@ -696,9 +713,7 @@ class TreeView:
             self.save_columns()
             return False
 
-        sort_column_id = column.get_sort_column_id()
-
-        if self._data_types[sort_column_id] == str or column.id in {"in_queue", "queue_position"}:
+        if self._data_types[sort_column_id] == str or column_id in {"in_queue", "queue_position"}:
             # String value (or queue position column): ascending sort by default
             first_sort_type = Gtk.SortType.ASCENDING
             second_sort_type = Gtk.SortType.DESCENDING
@@ -764,6 +779,15 @@ class TreeView:
         self._column_offsets[column_id] = offset
         self.save_columns()
 
+    def on_key_move_cursor(self, _widget, step, *_args):
+
+        if step != Gtk.MovementStep.BUFFER_ENDS:
+            return
+
+        # We are scrolling to the end using the End key. Disable the
+        # auto-scroll workaround to actually change the scroll adjustment value.
+        self._is_scrolling_to_row = True
+
     def on_v_adjustment_value(self, *_args):
 
         upper = self._v_adjustment.get_upper()
@@ -785,17 +809,18 @@ class TreeView:
         if not search_term:
             return True
 
-        for column_index in self._column_ids.values():
-            if model.get_column_type(column_index) != GObject.TYPE_STRING:
+        accepted_column_types = {"text", "number"}
+
+        for column_index, column_data in enumerate(self._columns.values()):
+            if "column_type" not in column_data:
                 continue
 
-            column_value = model.get_value(iterator, column_index).lower()
-
-            if column_value.startswith("nplus-"):
-                # Ignore icon name columns
+            if column_data["column_type"] not in accepted_column_types:
                 continue
 
-            if search_term.lower() in column_value:
+            column_value = model.get_value(iterator, column_index)
+
+            if column_value and search_term.lower() in column_value.lower():
                 return False
 
         return True
@@ -803,21 +828,10 @@ class TreeView:
     def on_tooltip(self, _widget, pos_x, pos_y, _keyboard_mode, tooltip):
 
         bin_x, bin_y = self.widget.convert_widget_to_bin_window_coords(pos_x, pos_y)
+        is_blank, path, column, _cell_x, _cell_y = self.widget.is_blank_at_pos(bin_x, bin_y)
 
-        # is_blank_at_pos() crashes in 4.12.0 - 4.12.2
-        # Remove check when GTK 4 version in Snap package is >= 4.12.3
-        if (4, 12, 2) >= (GTK_API_VERSION, GTK_MINOR_VERSION, GTK_MICRO_VERSION) >= (4, 12, 0):
-            result = self.widget.get_path_at_pos(bin_x, bin_y)
-
-            if not result:
-                return False
-
-            path, column, _cell_x, _cell_y = result
-        else:
-            is_blank, path, column, _cell_x, _cell_y = self.widget.is_blank_at_pos(bin_x, bin_y)
-
-            if is_blank:
-                return False
+        if is_blank:
+            return False
 
         iterator = self.model.get_iter(path)
 
@@ -861,32 +875,62 @@ class TreeView:
         clipboard.copy_text(value)
         return True
 
+    def on_start_search(self, *_args):
+        """Ctrl+F: start search."""
+
+        self.widget.emit("start-interactive-search")
+
+    def on_collapse_row_accelerator(self, *_args):
+        """Left: collapse row."""
+
+        iterator = self.get_focused_row()
+
+        if iterator is None:
+            return False
+
+        return self.collapse_row(iterator)
+
+    def on_collapse_row_blocked_accelerator(self, *_args):
+        """minus: collapse row (block search)."""
+
+        self.on_collapse_row_accelerator()
+        return True
+
+    def on_expand_row_accelerator(self, *_args):
+        """Right: expand row."""
+
+        iterator = self.get_focused_row()
+
+        if iterator is None:
+            return False
+
+        return self.expand_row(iterator)
+
+    def on_expand_row_blocked_accelerator(self, *_args):
+        """plus, equal: expand row (block search)."""
+
+        self.on_expand_row_accelerator()
+        return True
+
+    def on_expand_row_level_accelerator(self, *_args):
+        """\backslash: collapse or expand to show subs."""
+
+        iterator = self.get_focused_row()
+
+        if iterator is None:
+            return False
+
+        self.collapse_row(iterator)  # show 2nd level
+        self.expand_row(iterator)
+        return True
+
 
 # Legacy Functions (to be removed) #
 
 
-def verify_grouping_mode(mode):
-
-    # Map legacy values
-    if mode == "0":
-        mode = "ungrouped"
-
-    elif mode == "1":
-        mode = "folder_grouping"
-
-    elif mode == "2":
-        mode = "user_grouping"
-
-    # Verify mode validity
-    elif mode not in {"ungrouped", "folder_grouping", "user_grouping"}:
-        mode = "folder_grouping"
-
-    return mode
-
-
 def create_grouping_menu(window, active_mode, callback):
 
-    action_id = "grouping-" + "".join(random.choice(string.digits) for _ in range(8))
+    action_id = f"grouping-{GLib.uuid_string_random()}"
     menu = Gio.Menu()
 
     menuitem = Gio.MenuItem.new(_("Ungrouped"), f"win.{action_id}::ungrouped")
@@ -898,7 +942,7 @@ def create_grouping_menu(window, active_mode, callback):
     menuitem = Gio.MenuItem.new(_("Group by User"), f"win.{action_id}::user_grouping")
     menu.append_item(menuitem)
 
-    state = GLib.Variant("s", verify_grouping_mode(active_mode))
+    state = GLib.Variant("s", active_mode)
     action = Gio.SimpleAction(name=action_id, parameter_type=GLib.VariantType("s"), state=state)
     action.connect("change-state", callback)
 
