@@ -1,4 +1,4 @@
-# COPYRIGHT (C) 2020-2023 Nicotine+ Contributors
+# COPYRIGHT (C) 2020-2024 Nicotine+ Contributors
 #
 # GNU GENERAL PUBLIC LICENSE
 #    Version 3, 29 June 2007
@@ -36,6 +36,7 @@ class PortmapError(Exception):
 
 
 class BaseImplementation:
+    __slots__ = ("port", "local_ip_address")
 
     def __init__(self):
         self.port = None
@@ -52,6 +53,8 @@ class NATPMP(BaseImplementation):
     https://www.rfc-editor.org/rfc/rfc6886.
     """
 
+    __slots__ = ("_gateway_address",)
+
     NAME = "NAT-PMP"
     REQUEST_PORT = 5351
     REQUEST_ATTEMPTS = 2  # spec says 9, but 2 should be enough
@@ -59,6 +62,7 @@ class NATPMP(BaseImplementation):
     SUCCESS_RESULT = 0
 
     class PortmapResponse:
+        __slots__ = ("message", "result")
 
         def __init__(self, message):
 
@@ -77,6 +81,7 @@ class NATPMP(BaseImplementation):
             return self.message
 
     class PortmapRequest:
+        __slots__ = ("_public_port", "_private_port", "_lease_duration")
 
         RESERVED_VALUE = 0
         TCP_OP_CODE = 2
@@ -137,7 +142,7 @@ class NATPMP(BaseImplementation):
         else:
             gateway_pattern = re.compile(b"(?:default|0\\.0\\.0\\.0|::/0)\\s+([\\w\\.:]+)\\s+.*UG")
 
-        output = execute_command("netstat -rn", returnoutput=True)
+        output = execute_command("netstat -rn", returnoutput=True, hidden=True)
         return gateway_pattern.search(output).group(1)
 
     def _request_port_mapping(self, public_port, private_port, lease_duration):
@@ -190,6 +195,8 @@ class NATPMP(BaseImplementation):
 class UPnP(BaseImplementation):
     """Implementation of the UPnP protocol."""
 
+    __slots__ = ("_service",)
+
     NAME = "UPnP"
     MULTICAST_HOST = "239.255.255.250"
     MULTICAST_PORT = 1900
@@ -198,6 +205,8 @@ class UPnP(BaseImplementation):
     HTTP_REQUEST_TIMEOUT = 5
 
     class Service:
+        __slots__ = ("service_type", "control_url")
+
         def __init__(self, service_type, control_url):
             self.service_type = service_type
             self.control_url = control_url
@@ -205,6 +214,8 @@ class UPnP(BaseImplementation):
     class SSDPResponse:
         """Simple Service Discovery Protocol (SSDP) response."""
 
+        __slots__ = ("message", "headers")
+
         def __init__(self, message):
 
             import email.parser
@@ -218,6 +229,8 @@ class UPnP(BaseImplementation):
     class SSDPRequest:
         """Simple Service Discovery Protocol (SSDP) request."""
 
+        __slots__ = ("headers",)
+
         def __init__(self, search_target):
 
             self.headers = {
@@ -252,15 +265,21 @@ class UPnP(BaseImplementation):
             control_url = None
 
             try:
+                from urllib.error import HTTPError
                 from urllib.request import urlopen
                 from xml.etree import ElementTree
 
-                with urlopen(location_url, timeout=UPnP.HTTP_REQUEST_TIMEOUT) as response:
-                    response_body = response.read()
+                try:
+                    with urlopen(location_url, timeout=UPnP.HTTP_REQUEST_TIMEOUT) as response:
+                        response_body = response.read()
+
+                except HTTPError as error:
+                    # Received HTTP error, check what the response body says
+                    response_body = error.read()
 
                 log.add_debug("UPnP: Device description response from %s: %s", (location_url, response_body))
 
-                xml = ElementTree.fromstring(response_body.decode("utf-8"))
+                xml = ElementTree.fromstring(response_body.decode("utf-8", "replace"))
 
                 for service in xml.findall(".//{urn:schemas-upnp-org:device-1-0}service"):
                     found_service_type = service.find(".//{urn:schemas-upnp-org:device-1-0}serviceType").text
@@ -358,12 +377,13 @@ class UPnP(BaseImplementation):
                 while True:
                     try:
                         message = sock.recv(65507)  # Maximum size of UDP message
-                        UPnP.SSDP.add_service(services, locations, UPnP.SSDPResponse(message.decode("utf-8")))
+                        UPnP.SSDP.add_service(
+                            services, locations, UPnP.SSDPResponse(message.decode("utf-8", "replace")))
 
                     except socket.timeout:
                         break
 
-                log.add_debug("UPnP: %s service(s) detected", str(len(services)))
+                log.add_debug("UPnP: %s service(s) detected", len(services))
 
             return services
 
@@ -392,6 +412,7 @@ class UPnP(BaseImplementation):
         period of 12 hours.
         """
 
+        from urllib.error import HTTPError
         from urllib.request import Request
         from urllib.request import urlopen
         from xml.etree import ElementTree
@@ -431,10 +452,17 @@ class UPnP(BaseImplementation):
         log.add_debug("UPnP: Add port mapping request headers: %s", headers)
         log.add_debug("UPnP: Add port mapping request contents: %s", body)
 
-        with urlopen(Request(control_url, data=body, headers=headers), timeout=self.HTTP_REQUEST_TIMEOUT) as response:
-            response_body = response.read()
+        try:
+            request = Request(control_url, data=body, headers=headers)
+            with urlopen(request, timeout=self.HTTP_REQUEST_TIMEOUT) as response:
+                response_body = response.read()
 
-        xml = ElementTree.fromstring(response_body.decode("utf-8"))
+        except HTTPError as error:
+            # Received HTTP error, but response might also contain UPnP error code.
+            # E.g. MikroTik routers that send UPnP error 725 (OnlyPermanentLeasesSupported).
+            response_body = error.read()
+
+        xml = ElementTree.fromstring(response_body.decode("utf-8", "replace"))
 
         if xml.find(".//{http://schemas.xmlsoap.org/soap/envelope/}Body") is None:
             raise PortmapError(f"Invalid response: {response_body}")
@@ -463,11 +491,8 @@ class UPnP(BaseImplementation):
             raise PortmapError(_("No UPnP devices found"))
 
         # Perform the port mapping
-        log.add_debug("UPnP: Trying to redirect external WAN port %s TCP => %s port %s TCP", (
-            self.port,
-            self.local_ip_address,
-            self.port
-        ))
+        log.add_debug("UPnP: Trying to redirect external WAN port %s TCP => %s port %s TCP",
+                      (self.port, self.local_ip_address, self.port))
 
         error_code, error_description = self._request_port_mapping(
             public_port=self.port,
@@ -490,6 +515,7 @@ class UPnP(BaseImplementation):
         if not self._service:
             return
 
+        from urllib.error import HTTPError
         from urllib.request import Request
         from urllib.request import urlopen
 
@@ -521,14 +547,24 @@ class UPnP(BaseImplementation):
         log.add_debug("UPnP: Remove port mapping request headers: %s", headers)
         log.add_debug("UPnP: Remove port mapping request contents: %s", body)
 
-        with urlopen(
-                Request(control_url, data=body, headers=headers), timeout=self.HTTP_REQUEST_TIMEOUT) as response:
-            log.add_debug("UPnP: Remove port mapping response: %s", response.read())
+        try:
+            request = Request(control_url, data=body, headers=headers)
+            with urlopen(request, timeout=self.HTTP_REQUEST_TIMEOUT) as response:
+                response_body = response.read()
+
+        except HTTPError as error:
+            # Received HTTP error, but response body might contain useful information
+            response_body = error.read()
+
+        log.add_debug("UPnP: Remove port mapping response: %s", response_body)
 
 
 class PortMapper:
     """Class that handles Port Mapping."""
 
+    __slots__ = ("_active_implementation", "_has_port", "_is_mapping_port", "_timer",
+                 "_natpmp", "_upnp")
+
     RENEWAL_INTERVAL = 7200   # 2 hours
     LEASE_DURATION = 43200    # 12 hours
 
@@ -569,16 +605,15 @@ class PortMapper:
                 self._upnp.add_port_mapping(self.LEASE_DURATION)
 
             except Exception as upnp_error:
-                log.add_debug("UPnP not available, port forwarding failed: %s", upnp_error)
-
                 log.add(_("%(protocol)s: Failed to forward external port %(external_port)s: %(error)s"), {
                     "protocol": self._active_implementation.NAME,
                     "external_port": self._active_implementation.port,
                     "error": upnp_error
                 })
 
-                from traceback import format_exc
-                log.add_debug(format_exc())
+                if str(upnp_error) != _("No UPnP devices found"):
+                    from traceback import format_exc
+                    log.add_debug(format_exc())
 
                 self._active_implementation = None
                 self._is_mapping_port = False
@@ -606,10 +641,7 @@ class PortMapper:
             self._active_implementation.remove_port_mapping()
 
         except Exception as error:
-            log.add_debug("%(protocol)s: Failed to remove port mapping: %(error)s", {
-                "protocol": self._active_implementation.NAME,
-                "error": error
-            })
+            log.add_debug("%s: Failed to remove port mapping: %s", (self._active_implementation.NAME, error))
 
         self._active_implementation = None
         self._is_mapping_port = False
